#!/usr/bin/env python
"""
date2dec : Converts calendar dates into decimal dates.

This module was written by Arndt Piayda and then enhanced and
maintained by Matthias Cuntz while at Department of Computational
Hydrosystems, Helmholtz Centre for Environmental Research - UFZ,
Leipzig, Germany, and continued by Matthias Cuntz while at Institut
National de Recherche pour l'Agriculture, l'Alimentation et
l'Environnement (INRAE), Nancy, France.

Copyright (c) 2010-2020 Matthias Cuntz - mc (at) macu (dot) de
Released under the MIT License; see LICENSE file for details.

* Written Jun 2010 by Arndt Piayda
* Input can be scalar, list, array, or mix of it, Feb 2012, Matthias Cuntz
* Changed checks, added calendars decimal and decimal360, Feb 2012, Matthias Cuntz
* Changed units of proleptic_gregorian calendar from days since 0001-01-01 00:00:00 to days since 0001-01-00 00:00:00, Dec 2010, Matthias Cuntz
* Deal with Excel leap year error, Feb 2013, Matthias Cuntz
* Ported to Python 3, Feb 2013, Matthias Cuntz
* ascii/eng without time defaults to 00:00:00, Jul 2013, Matthias Cuntz
* Excel starts at 1 not at 0 on 01 January 1900 or 1904, Oct 2013, Matthias Cuntz
* Bug: 01.01.0001 was substracted if Julian calendar even with units given, Oct 2013, Matthias Cuntz
* Removed remnant of time treatment before time check in eng keyword, Nov 2013, Matthias Cuntz
* Adapted to new netCDF4/netcdftime (>= v1.0) and datetime (>= Python v2.7.9), Jun 2015, Matthias Cuntz
* Call date2num with list instead of single netCDF4.datetime objects, Oct 2015, Matthias Cuntz
* mo for months always integer, Oct 2016, Matthias Cuntz
* 00, 01, etc. for integers not accepted by Python 3, removed from examples and code, Nov 2016, Matthias Cuntz
* Using numpy docstring format, May 2020, Matthias Cuntz
* Succeed eng by en keyword as in ascii2ascii and dec2date, Jul 2020, Matthias Cuntz
* proleptic_gregorian instead of gregorian calendar for Excel dates, Jul 2020, Matthias Cuntz

.. moduleauthor:: Matthias Cuntz, Arndt Piayda

The following functions are provided

.. autosummary::
   date2dec
"""
from __future__ import division, absolute_import, print_function
from __future__ import division, absolute_import, print_function
from warnings import warn, filterwarnings
filterwarnings("default", category=DeprecationWarning)
import numpy as np


__all__ = ['date2dec']


def date2dec(calendar = 'standard', units=None,
             excelerr = True, yr=1,
             mo=1, dy=1, hr=0, mi=0, sc=0,
             ascii=None, en=None, eng=None):
    """
    Convert scalar and array_like with calendar dates into decimal
    dates. Supported calendar formats are standard, gregorian, julian,
    proleptic_gregorian, excel1900, excel1904, 365_day, noleap, 366_day,
    all_leap, 360_day, decimal, or decimal360.

    Input is year, month day, hour, minute, second or a combination of them.
    Input as date string is possible.

    Output is decimal date with day as unit.

    Parameters
    ----------
    yr : array_like, optional
        years (default: 1)
    mo : array_like, optional
        month (default: 1)
    dy : array_like, optional
        days (default: 1)
    hr : array_like, optional
        hours (default: 0)
    mi : array_like, optional
        minutes (default: 0)
    sc : array_like, optional
        seconds (default: 0)
    ascii : array_like, optional
        strings of the format 'dd.mm.yyyy hh:mm:ss'.
        Missing hour, minutes and/or seconds are set
        to their default values (0).

        `ascii` overwrites all other keywords.

        `ascii` and `eng` are mutually exclusive.
    en : array_like, optional
        strings of the format 'yyyy-mm-dd hh:mm:ss'.
        Missing hour, minutes and/or seconds are set
        to their default values (0).

        `en` overwrites all other keywords.

        `en` and `ascii` are mutually exclusive.
    eng : array_like, optional
        Same as en: obsolete.
    calendar : str, optional
        Calendar of output dates (default: 'standard').

        Possible values are:

        'standard', 'gregorian' = julian calendar from
        01.01.-4712 12:00:00 (BC) until 05.03.1583 00:00:00 and
        gregorian calendar from 15.03.1583 00:00:00 until now.
        Missing 10 days do not exsist.

        'julian' = julian calendar from 01.01.-4712 12:00:00 (BC)
         until now.

        'proleptic_gregorian' = gregorian calendar from
        01.01.0001 00:00:00 until now.

        'excel1900' = Excel dates with origin at
        01.01.1900 00:00:00.

        'excel1904' = Excel 1904 (Lotus) format.
        Same as excel1904 but with origin at
        01.01.1904 00:00:00.

        '365_day', 'noleap' = 365 days format,
        i.e. common years only (no leap years)
        with origin at 01.01.0001 00:00:00.

        '366_day', 'all_leap' = 366 days format,
        i.e. leap years only (no common years)
        with origin at 01.01.0001 00:00:00.

        '360_day' = 360 days format,
        i.e. years with only 360 days (30 days per month)
        with origin at 01.01.0001 00:00:00.

        'decimal' = decimal year instead of decimal days.

        'decimal360' = decimal year with a year of 360 days, i.e. 12 month with 30 days each.
    units : str, optional
        User set units of output dates. Must be a
        string in the format 'days since yyyy-mm-dd hh:mm:ss'.
        Default values are set automatically depending on `calendar`.
    excelerr : bool, optional
       In Excel, the year 1900 is normally considered a leap year,
       which it was not. By default, this error is taken into account
       if `calendar=='excel1900'` (default: True).

       1900 is not considered a leap year if `excelerr==False`.

    Returns
    -------
    array_like
       array_like with decimal dates. The type of output will be the same as the input type.

    Notes
    -----
    Most versions of `datetime` do not support negative years,
    i.e. Julian days < 1721423.5 = 01.01.0001 00:00:00.

    There is an issue in `netcdftime` version < 0.9.5 in proleptic_gregorian for dates before year 301:
    dec2date(date2dec(ascii='01.01.0300 00:00:00', calendar='proleptic_gregorian'), calendar='proleptic_gregorian')
    [300, 1, 2, 0, 0, 0]
    dec2date(date2dec(ascii='01.01.0301 00:00:00', calendar='proleptic_gregorian'), calendar='proleptic_gregorian')
    [301, 1, 1, 0, 0, 0]

    List input is only supported up to 2 dimensions.

    Requires `netcdftime.py` from module `netcdftime` available at:
    http://netcdf4-python.googlecode.com

    Examples
    --------
    # Some implementations of datetime do not support negative years
    >>> import datetime
    >>> if datetime.MINYEAR > 0:
    ...     print('The minimum year in your datetime implementation is ', datetime.MINYEAR)
    ...     print('i.e. it does not support negative years (BC).')
    The minimum year in your datetime implementation is  1
    i.e. it does not support negative years (BC).

    >>> if datetime.MINYEAR > 0:
    ...     year   = np.array([2000, 1810, 1630, 1510, 1271, 619, 2, 1])
    ... else:
    ...     year   = np.array([2000, 1810, 1630, 1510, 1271, 619, -1579, -4712])
    >>> month  = np.array([1, 4, 7, 9, 3, 8, 8, 1])
    >>> day    = np.array([5, 24, 15, 20, 18, 27, 23, 1])
    >>> hour   = np.array([12, 16, 10, 14, 19, 11, 20, 12])
    >>> minute = np.array([30, 15, 20, 35, 41, 8, 3, 0])
    >>> second = np.array([15, 10, 40, 50, 34, 37, 41, 0])
    >>> decimal = date2dec(calendar = 'standard', yr=year, mo=month, dy=day, hr=hour, mi=minute, sc=second)
    >>> nn = year.size
    >>> print('{:.14e} {:.14e} {:.14e} {:.14e}'.format(*decimal[:nn//2]))
    2.45154902100695e+06 2.38226217719907e+06 2.31660093101852e+06 2.27284810821759e+06
    >>> print('{:.14e} {:.14e}'.format(*decimal[nn//2:nn-2]))
    2.18536732053241e+06 1.94738596431713e+06
    >>> decimal = date2dec(calendar='standard', yr=year, mo=6, dy=15, hr=12, mi=minute, sc=second)
    >>> print('{:.14e} {:.14e} {:.14e} {:.14e}'.format(*decimal[:nn//2]))
    2.45171102100695e+06 2.38231401053241e+06 2.31657101435185e+06 2.27275102488426e+06
    >>> print('{:.14e} {:.14e}'.format(*decimal[nn//2:nn-2]))
    2.18545602886574e+06 1.94731300598380e+06

    # ascii input
    >>> if datetime.MINYEAR > 0:
    ...     a = np.array(['05.01.2000 12:30:15', '24.04.1810 16:15:10', '15.07.1630 10:20:40', '20.09.1510 14:35:50',
    ...                   '18.03.1271 19:41:34', '27.08. 619 11:08:37', '23.08.0002 20:03:41', '01.01.0001 12:00:00'])
    ... else:
    ...     a = np.array(['05.01.2000 12:30:15', '24.04.1810 16:15:10', '15.07.1630 10:20:40',  '20.09.1510 14:35:50',
    ...                   '18.03.1271 19:41:34', '27.08. 619 11:08:37', '23.08.-1579 20:03:41', '01.01.-4712 12:00:00'])
    >>> decimal = date2dec(calendar='standard', ascii=a)
    >>> nn = a.size
    >>> print('{:.14e} {:.14e} {:.14e} {:.14e}'.format(*decimal[:nn//2]))
    2.45154902100695e+06 2.38226217719907e+06 2.31660093101852e+06 2.27284810821759e+06
    >>> print('{:.14e} {:.14e}'.format(*decimal[nn//2:nn-2]))
    2.18536732053241e+06 1.94738596431713e+06

    # calendar = 'julian'
    >>> decimal = date2dec(calendar='julian', ascii=a)
    >>> print('{:.14e} {:.14e} {:.14e} {:.14e}'.format(*decimal[:nn//2]))
    2.45156202100695e+06 2.38227417719907e+06 2.31661093101852e+06 2.27284810821759e+06
    >>> print('{:.14e} {:.14e}'.format(*decimal[nn//2:nn-2]))
    2.18536732053241e+06 1.94738596431713e+06

    # calendar = 'proleptic_gregorian'
    >>> decimal = date2dec(calendar='proleptic_gregorian', ascii=a)
    >>> print('{:.7f} {:.7f} {:.7f} {:.7f}'.format(*decimal[:nn//2]))
    730123.5210069 660836.6771991 595175.4310185 551412.6082176
    >>> print('{:.7f} {:.7f}'.format(*decimal[nn//2:nn-2]))
    463934.8205324 225957.4643171

    # calendar = 'excel1900' WITH excelerr=True -> 1900 considered as leap year
    >>> d = np.array(['05.01.2000 12:30:15', '27.05.1950 16:25:10', '13.08.1910 10:40:55',
    ...               '01.03.1900 00:00:00', '29.02.1900 00:00:00', '28.02.1900 00:00:00',
    ...               '01.01.1900 00:00:00'])
    >>> decimal = date2dec(calendar='excel1900', ascii=d)
    >>> nn = d.size
    >>> print('{:.7f} {:.7f} {:.7f}'.format(*decimal[:nn//2]))
    36530.5210069 18410.6841435 3878.4450810
    >>> print('{:.1f} {:.1f} {:.1f} {:.1f}'.format(*decimal[nn//2:]))
    61.0 60.0 59.0 1.0

    # calendar = 'excel1900' WITH excelerr = False -> 1900 is NO leap year
    >>> decimal = date2dec(calendar='excel1900', ascii=d, excelerr=False)
    >>> print('{:.7f} {:.7f} {:.7f}'.format(*decimal[:nn//2]))
    36529.5210069 18409.6841435 3877.4450810
    >>> print('{:.1f} {:.1f} {:.1f} {:.1f}'.format(*decimal[nn//2:]))
    60.0 60.0 59.0 1.0

    # calendar = 'excel1904'
    >>> decimal = date2dec(calendar='excel1904', ascii=d[:nn//2])
    >>> print('{:.7f} {:.7f} {:.7f}'.format(*decimal[:nn//2]))
    35069.5210069 16949.6841435 2417.4450810

    # calendar = '365_day'
    >>> g = np.array(['18.08.1972 12:30:15', '25.10.0986 12:30:15', '28.11.0493 22:20:40', '01.01.0001 00:00:00'])
    >>> decimal = date2dec(calendar='365_day', ascii=g)
    >>> nn = g.size
    >>> print('{:.7f} {:.7f} {:.7f} {:.7f}'.format(*decimal[:nn]))
    719644.5210069 359822.5210069 179911.9310185 0.0000000

    # calendar = '366_day'
    >>> decimal = date2dec(calendar='366_day', ascii=g)
    >>> print('{:.7f} {:.7f} {:.7f} {:.7f}'.format(*decimal[:nn]))
    721616.5210069 360808.5210069 180404.9310185 0.0000000

    # 360_day does not work with netcdftime.py version equal or below 0.9.2
    # calendar = '360_day'
    >>> decimal = date2dec(calendar='360_day', ascii=g)
    >>> print('{:.7f} {:.7f} {:.7f} {:.7f}'.format(*decimal[:nn]))
    709787.5210069 354894.5210069 177447.9310185 0.0000000

    >>> print('{:.7f}'.format(date2dec(yr=1992, mo=1, dy=26, hr=2, mi=0, sc=0, calendar='decimal')))
    1992.0685337
    >>> print('{:.7f}'.format(date2dec(ascii='26.01.1992 02:00', calendar='decimal360')))
    1992.0696759
    >>> print('{:.7f} {:.7f}'.format(*date2dec(ascii=['26.01.1992 02:00','26.01.1992 02:00'], calendar='decimal360')))
    1992.0696759 1992.0696759
    >>> print('{:.7f} {:.7f}'.format(*date2dec(yr=[1992,1992], mo=1, dy=26, hr=2, mi=0, sc=0, calendar='decimal360')))
    1992.0696759 1992.0696759
    >>> print('{:.7f} {:.7f}'.format(*date2dec(yr=np.array([1992,1992]), mo=1, dy=26, hr=2, mi=0, sc=0, calendar='decimal360')))
    1992.0696759 1992.0696759
    >>> decimal = date2dec(ascii=[['26.01.1992 02:00','26.01.1992 02:00'],
    ...                           ['26.01.1992 02:00','26.01.1992 02:00'],
    ...                           ['26.01.1992 02:00','26.01.1992 02:00']],
    ...                    calendar='decimal360')
    >>> print('{:.7f} {:.7f}'.format(*decimal[0]))
    1992.0696759 1992.0696759
    >>> print('{:.7f} {:.7f}'.format(*decimal[2]))
    1992.0696759 1992.0696759
    >>> print((date2dec(ascii='01.03.2003 00:00:00') - date2dec(ascii='01.03.2003')) == 0.)
    True

    # en
    >>> decimal = date2dec(en='1992-01-26 02:00', calendar='decimal360')
    >>> print('{:.7f}'.format(decimal))
    1992.0696759
    >>> decimal = date2dec(eng='1992-01-26 02:00', calendar='decimal360')
    >>> print('{:.7f}'.format(decimal))
    1992.0696759

    History
    -------
    Written  Arndt Piayda, Jun 2010
    Modified Matthias Cuntz, Feb 2012 - All input can be scalar, list or array, also a mix
                                        - Changed checks for easier extension
                                        - decimal, decimal360
             Matthias Cuntz, Dec 2012 - change unit of proleptic_gregorian
                                        from 'days since 0001-01-01 00:00:00'
                                        to   'days since 0001-01-00 00:00:00'
             Matthias Cuntz, Feb 2013 - solved Excel leap year problem.
             Matthias Cuntz, Feb 2013 - ported to Python 3
             Matthias Cuntz, Jul 2013 - ascii/eng without time defaults to 00:00:00
             Matthias Cuntz, Oct 2013 - Excel starts at 1 not at 0
             Matthias Cuntz, Oct 2013 - units bugs, e.g. 01.01.0001 was substracted if Julian calendar even with units
             Matthias Cuntz, Nov 2013 - removed remnant of time treatment before time check in eng keyword
             Matthias Cuntz, Jun 2015 - adapted to new netCDF4/netcdftime (>= v1.0) and datetime (>= Python v2.7.9)
             Matthias Cuntz, Oct 2015 - call date2num with list instead of single netCDF4.datetime objects
             Matthias Cuntz, Oct 2016 - netcdftime provided even with netCDF4 > 1.0.0; make mo for months always integer
             Matthias Cuntz, Nov 2016 - 00, 01, etc. for integers not accepted by Python3
             Matthias Cuntz, May 2020 - numpy docstring format
             Matthias Cuntz, Jul 2020 - en for eng
             Matthias Cuntz, Jul 2020 - use proleptic_gregorian for Excel dates
             Matthias Cuntz, May 2021 - rm np.int and np.float
                                      - (almost) flake8 compatible
    """
    warn('The function date2dec is deprecated and superseded by'
         ' date2num/date2dec.', category=DeprecationWarning)
    #
    # Checks
    calendars = ['standard', 'gregorian', 'julian', 'proleptic_gregorian',
                 'excel1900', 'excel1904', '365_day', 'noleap', '366_day',
                 'all_leap', '360_day', 'decimal', 'decimal360']
    # add little epsilon to julian dates for better reconversion
    muleps = np.finfo(float).eps
    try:
        import cftime as nt
        if (nt.__version__ <= '1.0.4'):
            muleps = 0.  # before 1.0.5, cftime used eps
    except:
        import netCDF4 as nt
    try:
        tst = nt.date2num
        tst = nt.datetime
    except:
        try:
            import netcdftime as nt
            if ((nt.__version__ <= '0.9.2') & (calendar == '360_day')):
                raise ValueError(
                    "date2dec error: Your version of netcdftime.py is equal"
                    " or below 0.9.2. The 360_day calendar does not work with"
                    " arrays here. Please download a newer one.")
        except:
            raise ImportError('Could not determine netcdf time library.')
    #
    calendar = calendar.lower()
    if (calendar not in calendars):
        raise ValueError(
            "date2dec error: Wrong calendar!" +
            " Choose: " + ''.join([i + ' ' for i in calendars]))
    # obsolete eng
    if (eng is not None):
        if (en is not None):
            raise ValueError("date2dec error: 'eng' was succeeded by 'en'. Only one can be given.")
        else:
            en = eng
    # if ascii input is given by user, other input will be neglected
    # calculation of input size and shape
    if (ascii is not None) and (en is not None):
        raise ValueError("date2dec error: 'ascii' and 'en' mutually exclusive")
    if (ascii is not None):
        islist = type(ascii) != type(np.array(ascii))
        isarr = np.ndim(ascii)
        if (islist & (isarr > 2)):
            raise ValueError("date2dec error: ascii input is list > 2D; Use array input")
        if isarr == 0:
            ascii = np.array([ascii])
        else:
            ascii = np.array(ascii)
        insize   = ascii.size
        outsize  = insize
        outshape = ascii.shape
        asciifl  = ascii.flatten()
        timeobj  = np.zeros(insize, dtype=object)
        # slicing of ascii strings to implement in datetime object. missing times
        # will be set to 0.
        yr = np.zeros(insize, dtype=int)
        mo = np.zeros(insize, dtype=int)
        dy = np.zeros(insize, dtype=int)
        hr = np.zeros(insize, dtype=int)
        mi = np.zeros(insize, dtype=int)
        sc = np.zeros(insize, dtype=int)
        for i in range(insize):
            aa      = asciifl[i].split('.')
            dy[i]   = int(aa[0])
            mo[i]   = int(aa[1])
            tail    = aa[2].split()
            yr[i]   = int(tail[0])
            if len(tail) > 1:
                tim     = tail[1].split(':')
                hr[i]   = int(tim[0])
                if len(tim) > 1:
                    mi[i] = int(tim[1])
                else:
                    mi[i] = 0
                if len(tim) > 2:
                    sc[i] = int(tim[2])
                else:
                    sc[i] = 0
            else:
                hr[i] = 0
                mi[i] = 0
                sc[i] = 0
            if ((yr[i] == 1900) & (mo[i] == 2) & (dy[i] == 29)):
                timeobj[i] = nt.datetime(yr[i], 3, 1, hr[i], mi[i], sc[i])
            else:
                timeobj[i] = nt.datetime(yr[i], mo[i], dy[i], hr[i], mi[i],
                                         sc[i])
    if (en is not None):
        islist = type(en) != type(np.array(en))
        isarr  = np.ndim(en)
        if isarr == 0:
            en = np.array([en])
        else:
            en = np.array(en)
        if (islist & (isarr > 2)):
            raise ValueError("date2dec error: en input is list > 2D; Use array input")
        en = np.array(en)
        insize   = en.size
        outsize  = insize
        outshape = en.shape
        enfl     = en.flatten()
        timeobj  = np.zeros(insize, dtype=object)
        # slicing of en strings to implement in datetime object. missing times
        # will be set to 0.
        yr = np.zeros(insize, dtype=int)
        mo = np.zeros(insize, dtype=int)
        dy = np.zeros(insize, dtype=int)
        hr = np.zeros(insize, dtype=int)
        mi = np.zeros(insize, dtype=int)
        sc = np.zeros(insize, dtype=int)
        for i in range(insize):
            ee      = enfl[i].split('-')
            yr[i]   = int(ee[0])
            mo[i]   = int(ee[1])
            tail    = ee[2].split()
            dy[i]   = int(tail[0])
            if len(tail) > 1:
                tim     = tail[1].split(':')
                hr[i]   = int(tim[0])
                if len(tim) > 1:
                    mi[i] = int(tim[1])
                else:
                    mi[i] = 0
                if len(tim) > 2:
                    sc[i] = int(tim[2])
                else:
                    sc[i] = 0
            else:
                hr[i] = 0
                mi[i] = 0
                sc[i] = 0
            if ((yr[i] == 1900) & (mo[i] == 2) & (dy[i] == 29)):
                timeobj[i] = nt.datetime(yr[i], 3, 1, hr[i], mi[i], sc[i])
            else:
                timeobj[i] = nt.datetime(yr[i], mo[i], dy[i], hr[i], mi[i],
                                         sc[i])
    # if no ascii input, other inputs will be concidered
    # calculation of input sizes, shapes and number of axis
    if (ascii is None) and (en is None):
        isnlist1 = type(yr) == type(np.array(yr))
        isarr1   = np.ndim(yr)
        if isarr1 == 0:
            yr = np.array([yr])
        else:
            yr = np.array(yr)
        isnlist2 = type(mo) == type(np.array(mo))
        isarr2   = np.ndim(mo)
        if isarr2 == 0:
            mo = np.array([mo], dtype=int)
        else:
            mo = np.array(mo, dtype=int)
        isnlist3 = type(dy) == type(np.array(dy))
        isarr3   = np.ndim(dy)
        if isarr3 == 0:
            dy = np.array([dy])
        else:
            dy = np.array(dy)
        isnlist4 = type(hr) == type(np.array(hr))
        isarr4   = np.ndim(hr)
        if isarr4 == 0:
            hr = np.array([hr])
        else:
            hr = np.array(hr)
        isnlist5 = type(mi) == type(np.array(mi))
        isarr5   = np.ndim(mi)
        if isarr5 == 0:
            mi = np.array([mi])
        else:
            mi = np.array(mi)
        isnlist6 = type(sc) == type(np.array(sc))
        isarr6   = np.ndim(sc)
        if isarr6 == 0:
            sc = np.array([sc])
        else:
            sc = np.array(sc)
        islist = not (isnlist1 | isnlist2 | isnlist3 | isnlist4 | isnlist5 |
                      isnlist6)
        isarr  = isarr1 + isarr2 + isarr3 + isarr4 + isarr5 + isarr6
        shapes = [np.shape(yr), np.shape(mo), np.shape(dy), np.shape(hr),
                  np.shape(mi), np.shape(sc)]
        nyr    = np.size(yr)
        nmo    = np.size(mo)
        ndy    = np.size(dy)
        nhr    = np.size(hr)
        nmi    = np.size(mi)
        nsc    = np.size(sc)
        sizes  = [nyr, nmo, ndy, nhr, nmi, nsc]
        nmax   = np.amax(sizes)
        ii     = np.argmax(sizes)
        outshape = shapes[ii]
        if (islist & (np.size(outshape) > 2)):
            raise ValueError("date2dec error: input is list > 2D; Use array input.")
        if nyr < nmax:
            if nyr == 1:
                yr  = np.ones(outshape, dtype=int) * yr
            else:
                raise ValueError("date2dec error: size of yr != max input or 1.")
        if nmo < nmax:
            if nmo == 1:
                mo  = np.ones(outshape, dtype=int) * mo
            else:
                raise ValueError("date2dec error: size of mo != max input or 1.")
        if ndy < nmax:
            if ndy == 1:
                dy  = np.ones(outshape, dtype=int) * dy
            else:
                raise ValueError("date2dec error: size of dy != max input or 1.")
        if nhr < nmax:
            if nhr == 1:
                hr  = np.ones(outshape, dtype=int) * hr
            else:
                raise ValueError("date2dec error: size of hr != max input or 1.")
        if nmi < nmax:
            if nmi == 1:
                mi  = np.ones(outshape, dtype=int) * mi
            else:
                raise ValueError("date2dec error: size of mi != max input or 1.")
        if nsc < nmax:
            if nsc == 1:
                sc  = np.ones(outshape, dtype=int) * sc
            else:
                raise ValueError("date2dec error: size of sc != max input or 1.")
        indate  = [yr, mo, dy, hr, mi, sc]
        insize  = [np.size(i) for i in indate]
        inshape = [np.shape(i) for i in indate]
        dimnum  = [len(i) for i in inshape]
        # calculation of maximum input size and maximum number of axis for
        # reshaping the output
        indate  = [i.flatten() for i in indate]
        outsize = max(insize)
        timeobj = np.zeros(outsize, dtype=object)
        # datetime object is constructed
        for i in range(outsize):
            iyr = int(indate[0][i])
            imo = int(indate[1][i])
            idy = int(indate[2][i])
            ihr = int(indate[3][i])
            imi = int(indate[4][i])
            isc = int(indate[5][i])
            if ((iyr == 1900) & (imo == 2) & (idy == 29)):
                timeobj[i] = nt.datetime(iyr, 3, 1, ihr, imi, isc)
            else:
                timeobj[i] = nt.datetime(iyr, imo, idy, ihr, imi, isc)
    # depending on chosen calendar and optional set of the time units
    # decimal date is calculated
    output = np.zeros(outsize)
    try:
        t0 = nt.datetime(1582, 10, 4, 23, 59, 59, 999999)
    except:
        t0 = nt.datetime(1582, 10, 4, 23, 59, 59)
    t1    = nt.datetime(1582, 10, 15, 0, 0, 0)
    is121 = True if (min(timeobj) <= t0) and (max(timeobj) >= t1) else False
    if (calendar == 'standard') or (calendar == 'gregorian'):
        if not units:
            units = 'days since 0001-01-01 12:00:00'
            dec0 = 1721424
        else:
            dec0 = 0
        if is121 and (nt.__version__ < '1.2.2'):
            for ii, tt in enumerate(timeobj):
                output[ii] = nt.date2num(tt, units,
                                         calendar='gregorian') + dec0
        else:
            output = nt.date2num(timeobj, units,
                                 calendar='gregorian').astype(float) + dec0
        output += np.abs(output) * muleps
    elif calendar == 'julian':
        if not units:
            units = 'days since 0001-01-01 12:00:00'
            dec0 = 1721424
        else:
            dec0 = 0
        if is121 and (nt.__version__ < '1.2.2'):
            for ii, tt in enumerate(timeobj):
                output[ii] = nt.date2num(tt, units,
                                         calendar='julian') + dec0
        else:
            output = nt.date2num(timeobj, units,
                                 calendar='julian').astype(float) + dec0
        output += np.abs(output) * muleps
    elif calendar == 'proleptic_gregorian':
        if not units:
            units = 'days since 0001-01-01 00:00:00'
        if is121 and (nt.__version__ < '1.2.2'):
            for ii, tt in enumerate(timeobj):
                output[ii] = nt.date2num(tt, units,
                                         calendar='proleptic_gregorian')
        else:
            output = nt.date2num(timeobj, units,
                                 calendar='proleptic_gregorian').astype(float)
        output += np.abs(output) * muleps
    elif calendar == 'excel1900':
        doerr = False
        if not units:
            units = 'days since 1899-12-31 00:00:00'
            if excelerr:
                doerr = True
        if is121 and (nt.__version__ < '1.2.2'):
            for ii, tt in enumerate(timeobj):
                output[ii] = nt.date2num(tt, units,
                                         calendar='proleptic_gregorian')
        else:
            output = nt.date2num(timeobj, units,
                                 calendar='proleptic_gregorian').astype(float)
        if doerr:
            output = np.where(output >= 60., output + 1., output)
            # date2num treats 29.02.1900 as 01.03.1990,
            # i.e. is the same decimal number
            if np.any((output >= 61.) & (output < 62.)):
                for i in range(outsize):
                    # if (timeobj[i].year==1900) & (timeobj[i].month==2) &
                    #    (timeobj[i].day==29):
                    #     output[i] -= 1.
                    if (yr[i] == 1900) & (mo[i] == 2) & (dy[i] == 29):
                        output[i] -= 1.
        output += np.abs(output) * muleps
    elif calendar == 'excel1904':
        if not units:
            units = 'days since 1903-12-31 00:00:00'
        if is121 and (nt.__version__ < '1.2.2'):
            for ii, tt in enumerate(timeobj):
                output[ii] = nt.date2num(tt, units,
                                         calendar='proleptic_gregorian')
        else:
            output = nt.date2num(timeobj, units,
                                 calendar='proleptic_gregorian').astype(float)
        output += np.abs(output) * muleps
    elif (calendar == '365_day') or (calendar == 'noleap'):
        if not units:
            units = 'days since 0001-01-01 00:00:00'
        if is121 and (nt.__version__ < '1.2.2'):
            for ii, tt in enumerate(timeobj):
                output[ii] = nt.date2num(tt, units, calendar='365_day')
        else:
            output = nt.date2num(timeobj, units,
                                 calendar='365_day').astype(float)
        output += np.abs(output) * muleps
    elif (calendar == '366_day') or (calendar == 'all_leap'):
        if not units:
            units = 'days since 0001-01-01 00:00:00'
        if is121 and (nt.__version__ < '1.2.2'):
            for ii, tt in enumerate(timeobj):
                output[ii] = nt.date2num(tt, units, calendar='366_day')
        else:
            output = nt.date2num(timeobj, units,
                                 calendar='366_day').astype(float)
        output += np.abs(output) * muleps
    elif calendar == '360_day':
        if not units:
            units = 'days since 0001-01-01 00:00:00'
        if is121 and (nt.__version__ < '1.2.2'):
            for ii, tt in enumerate(timeobj):
                output[ii] = nt.date2num(tt, units, calendar='360_day')
        else:
            output = nt.date2num(timeobj, units,
                                 calendar='360_day').astype(float)
        output += np.abs(output) * muleps
    elif calendar == 'decimal':
        ntime = np.size(yr)
        leap  = np.array((((yr % 4) == 0) & ((yr % 100) != 0)) |
                         ((yr % 400) == 0)).astype(int)
        tdy   = np.array(dy, dtype=float)
        diy   = np.array([ [-9, 0, 31, 59,  90, 120, 151, 181, 212,
                            243, 273, 304, 334, 365],
                           [-9, 0, 31, 60,  91, 121, 152, 182, 213,
                            244, 274, 305, 335, 366] ])
        for i in range(ntime):
            tdy[i] = tdy[i] + np.array(diy[leap[i], mo[i]], dtype=float)
        days_year = 365.
        output    = ( np.array(yr, dtype=float) +
                      ((tdy - 1.) * 24. + np.array(hr, dtype=float) +
                       np.array(mi, dtype=float) / 60. +
                       np.array(sc, dtype=float) / 3600.) /
                      ((days_year + np.array(leap, dtype=float)) * 24.) )
        # for numerical stability, i.e. back and forth transforms
        output += 1e-08  # 1/3 sec
    elif calendar == 'decimal360':
        ntime = np.size(yr)
        tdy   = np.array(dy, dtype=float)
        diy   = np.array([-9,  0, 30, 60, 90, 120, 150, 180, 210, 240, 270,
                          300, 330, 360])
        for i in range(ntime):
            tdy[i] = tdy[i] + np.array(diy[mo[i]], dtype=float)
        days_year = 360.
        output    = ( np.array(yr, dtype=float) +
                      ((tdy - 1.) * 24. + np.array(hr, dtype=float) +
                       np.array(mi, dtype=float) / 60. +
                       np.array(sc, dtype=float) / 3600.) /
                      (days_year * 24.) )
        # for numerical stability, i.e. back and forth transforms
        output += 1e-08  # 1/3 sec
    else:
        raise ValueError("date2dec error: calendar not implemented; should have been catched before.")

    # return of reshaped output
    output = np.reshape(output, outshape)
    if isarr == 0:
        output = float(output)
    else:
        if islist:
            ns = np.size(outshape)
            if ns == 1:
                output = [i for i in output]
            else:
                loutput = [ i for i in output[:, 0]]
                for i in range(np.size(output[:, 0])):
                    loutput[i] = list(np.squeeze(output[i, :]))
                output = loutput

    return output


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
"""
netCDF4 functions to a copy netcdf file while doing some transformations on
variables and dimensions.

This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

It borrows the idea of get_variable_definition from the netcdf4 thin layer of
David Schaefer.

:copyright: Copyright 2020-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided

.. autosummary::
   copy_dimensions
   copy_file
   copy_global_attributes
   copy_variables
   create_new_variable
   create_variables
   get_fill_value_for_dtype
   get_variable_definition
   set_output_filename

History
    * Written Apr 2020 by Matthias Cuntz (mc (at) macu (dot) de)
    * Added get_fill_value_for_dtype, Mar 2021, Matthias Cuntz
    * Added create_new_variable, Mar 2021, Matthias Cuntz
    * flake8 compatible, Mar 2021, Matthias Cuntz
    * Remove keyword in copy_global_attributes, Mar 2021, Matthias Cuntz
    * Add timedim in create_variables, Dec 2021, Matthias Cuntz
    * Rename functions, e.g. create_dimensions -> copy_dimensions,
      May 2022, Matthias Cuntz
    * Add copy_variables, May 2022, Matthias Cuntz
    * Delete unnecessary HDF5 filters in variable definition for compatibility
      with netcdf4 > 1.6.0, Jun 2022, Matthias Cuntz
    * Make get_variable_definition public, Apr 2023, Matthias Cuntz
    * get_variable_definition returns _FillValue and only then missing_value,
      Jul 2023, Matthias Cuntz
    * create_variables sets missing_value attribute if present even if used for
      _FillValue, Jul 2023, Matthias Cuntz

"""
import numpy as np
import netCDF4 as nc


__all__ = ['copy_dimensions', 'copy_file', 'copy_global_attributes',
           'copy_variables', 'create_new_variable', 'create_variables',
           'get_fill_value_for_dtype', 'get_variable_definition',
           'set_output_filename']


def _tolist(arg):
    """
    Assure that *arg* is a list, e.g. if string or None are given.

    Parameters
    ----------
    arg :
        Argument to make list

    Returns
    -------
    list
        list(arg)

    Examples
    --------
    >>> _tolist('string')
    ['string']
    >>> _tolist([1,2,3])
    [1, 2, 3]
    >>> _tolist(None)
    [None]

    """
    if isinstance(arg, str):
        return [arg]
    try:
        return list(arg)
    except TypeError:
        return [arg]


def get_variable_definition(ncvar):
    """
    Collect information on input variable.

    Parameters
    ----------
    ncvar : netcdf4 variable
        Variable of input file

    Returns
    -------
    dict
        Containing information on input variable withkey/value pairs.
        The following keys are returned: 'name', 'dtype', 'dimensions',
        'fill_vallue', 'chunksizes'

    Examples
    --------
    .. code-block:: python

       get_variable_definition(fi.variables['GPP'])

    """
    out = ncvar.filters() if ncvar.filters() else {}
    # Delete HDF5 filters that are False, e.g. because the plugin is not
    # available. zlib is always available.
    # Necessary for netcdf4 > 1.6.0 because it gives all possible filters,
    # installed or not, i.e. 'zlib', 'szip', 'zstd', 'bzip2', 'blosc',
    # 'shuffle', 'complevel', 'fletcher32'
    todel = [ oo for oo in out if (oo != 'zlib') and (not out[oo]) ]
    for oo in todel:
        del out[oo]
    # chunksizes
    chunks = None
    if "chunking" in dir(ncvar):
        if ncvar.chunking() is not None:
            if not isinstance(ncvar.chunking(), str):
                chunks = list(ncvar.chunking())
    # missing value
    if "_FillValue" in dir(ncvar):
        ifill = ncvar._FillValue
    elif "missing_value" in dir(ncvar):
        ifill = ncvar.missing_value
    else:
        ifill = None
    # output variable
    out.update({
        "name":       ncvar.name,
        "dtype":      ncvar.dtype,
        "dimensions": list(ncvar.dimensions),
        "fill_value": ifill,
        "chunksizes": chunks,
    })
    return out


def copy_file(ifile, ofile, timedim='time',
              removevar=[], renamevar={}, replacevar={}, replaceatt={},
              noclose=False):
    """
    Copy variables from input file into output file.

    Parameters
    ----------
    ifile : str
        File name of netcdf input file
    ofile : str
        File name of netcdf output file
    timedim : str, optional
        Name of time dimension in input file (default: 'time').
    removevar : list of str, optional
        Do not copy variables given in *removevar* to output file.
    renamevar : dict, optional
        Copy variables from input file with different name in output file.
        Variable names in input file are given as dictionary keys,
        corresponding variable names of output file are give as dictionary
        values.
    replacevar : dict, optional
        Replace existing variables with variables in dictionary.
        Variable names in input file are given as dictionary keys,
        dictionary values are also dictionaries where keys are the output
        variable name and values are the output variable values.
    replaceatt : dict, optional
        Replace or set attributes of variables in dictionary keys
        (case sensitive). Dictionary values are also dictionaries with
        {'attribute_name': attribute_value}. Dictionary keys are the names
        of the output variables after renaming and replacing.
    noclose : bool, optional
        Return file handle of opened output file for further manipulation
        if True (default: False)

    Returns
    -------
    nothing or file_handle
        The output file will have the altered or unaltered variables
        copied from the input file.

    Examples
    --------
    .. code-block:: python

       ovar = np.arange(100)
       copy_variable('in.nc', 'out.nc',
                     renamevar={'lon': 'longitude'},
                     replacevar={'var1': {'arange': ovar}},
                     replaceatt={'arange':
                                 {'long_name': 'A range', 'unit': '-'}})

    """
    import time as ptime
    fi = nc.Dataset(ifile, 'r')
    if 'file_format' in dir(fi):
        fo = nc.Dataset(ofile, 'w', format=fi.file_format)
    else:  # pragma: no cover
        fo = nc.Dataset(ofile, 'w', format='NETCDF4')

    # meta data
    copy_global_attributes(
        fi, fo, add={'history': ptime.asctime() + ': ' +
                     'copy_variables(' + ifile + ', ' + ofile + ')'})

    # copy dimensions
    copy_dimensions(fi, fo)

    # create variables
    # rename replace variables as well
    xreplacevar = {}
    for rr in replacevar:
        nn = replacevar[rr].copy()
        kk = nn.popitem()[0]
        xreplacevar.update({rr: kk})
    arenamevar = renamevar.copy()
    arenamevar.update(xreplacevar)
    # create static variables (independent of time)
    create_variables(fi, fo, time=False, timedim=timedim, fill=True,
                     removevar=removevar, renamevar=arenamevar)
    # create dynamic variables (time dependent)
    create_variables(fi, fo, time=True, timedim=timedim, fill=True,
                     removevar=removevar, renamevar=arenamevar)

    # set extra attributes
    for rr in replaceatt:
        ovar = fo.variables[rr]
        att = replaceatt[rr]
        for aa in att:
            ovar.setncattr(aa, att[aa])

    # copy variables
    # do not copy replace variables
    aremovevar = _tolist(removevar)
    aremovevar.extend(xreplacevar.keys())
    copy_variables(fi, fo, time=False, timedim=timedim,
                   removevar=aremovevar, renamevar=renamevar)
    copy_variables(fi, fo, time=True, timedim=timedim,
                   removevar=aremovevar, renamevar=renamevar)

    # set replace variables
    for rr in replacevar:
        odict = replacevar[rr].copy()
        oname, oval = odict.popitem()
        ovar = fo.variables[oname]
        ovar[:] = oval

    fi.close()
    if noclose:
        return fo
    else:
        fo.close()
        return


def copy_dimensions(fi, fo, removedim=[], renamedim={}, changedim={},
                    adddim={}):
    """
    Create dimensions in output file from dimensions in input file.

    Parameters
    ----------
    fi : file_handle
        File handle of opened netcdf input file
    fo : file_handle
        File handle of opened netcdf output file
    removedim : list of str, optional
        Do not create dimensions given in *removedim* in output file.
    renamedim : dict, optional
        Rename dimensions in output file compared to input file.
        Dimension names in input file are given as dictionary keys,
        corresponding dimension names of output file are give as dictionary
        values.
    changedim : dict, optional
        Change the size of the output dimension compared to the input file.
        Dimension names are given as dictionary keys, corresponding dimension
        sizes are given as dictionary values.
    adddim : dict, optional
        Add dimension to output file.
        New dimension names are given as dictionary keys and new dimension
        sizes are given as dictionary values.

    Returns
    -------
    nothing
        The output file will have the altered and unaltered dimensions
        of the input file.

    Examples
    --------
    .. code-block:: python

       copy_dimensions(fi, fo, removedim=['patch'],
                       renamedim={'x': 'lon', 'y': 'lat'},
                       changedim={'mland': 1})

    """
    removedim = _tolist(removedim)
    for d in fi.dimensions.values():
        # remove dimension if in removedim
        if d.name not in removedim:
            # change dimension size if in changedim
            if d.name in changedim.keys():
                nd = changedim[d.name]
            elif d.isunlimited():
                nd = None
            else:
                nd = len(d)
            # rename dimension if in renamedim
            if d.name in renamedim.keys():
                oname = renamedim[d.name]
            else:
                oname = d.name
            # create dimension
            fo.createDimension(oname, nd)
    # add new dimensions
    for d in adddim.keys():
        if d not in fo.dimensions:
            fo.createDimension(d, adddim[d])
    return


def copy_global_attributes(fi, fo, add={}, remove=[]):
    """
    Create global output file attributes from input global file attributes.

    Parameters
    ----------
    fi : file_handle
        File handle of opened netcdf input file
    fo : file_handle
        File handle of opened netcdf output file
    add : dict, optional
        dict values will be given to attributes given in dict keys.
        Attributes will be created if they do not exist yet.
    remove : list, optional
        Do not create global attributes given in *remove* in the output file.

    Returns
    -------
    nothing
        Output will have global file attributes

    Examples
    --------
    .. code-block:: python

       copy_global_attributes(
           fi, fo, add={'history': time.asctime()+': '+' '.join(sys.argv)})

    """
    for k in fi.ncattrs():
        if k not in remove:
            iattr = fi.getncattr(k)
            # add to existing global attribute
            if k in add.keys():
                iattr += '\n' + add[k]
            fo.setncattr(k, iattr)
    # add if global attribute does not exist yet
    for k in add.keys():
        if k not in fi.ncattrs():
            fo.setncattr(k, add[k])
    return


def copy_variables(fi, fo, time=None, timedim='time',
                   removevar=[], renamevar={}):
    """
    Copy variables from input file into output file.

    Parameters
    ----------
    fi : file_handle
        File handle of opened netcdf input file
    fo : file_handle
        File handle of opened netcdf output file
    time : None or bool, optional
        None:  copy all variables (default).
        True:  copy only variables having dimension *timedim*.
        False: copy only variables that do not have dimension *timedim*.
    timedim : str, optional
        Name of time dimension (default: 'time').
    removevar : list of str, optional
        Do not copy variables given in *removevar* to output file.
    renamevar : dict, optional
        Copy variables from input file with different name in output file.
        Variable names in input file are given as dictionary keys,
        corresponding variable names of output file are give as dictionary
        values.

    Returns
    -------
    nothing
        The output file will have the altered or unaltered variables
        copied from the input file.

    Examples
    --------
    .. code-block:: python

       copy_variable(fi, fo, fill=True, renamevar={'lon': 'longitude'})

    """
    removevar = _tolist(removevar)
    # just copy all variables if no time dimension or
    # time dimension not known
    # Could be improved by looking for the unlimited dimension
    if timedim not in fi.dimensions:
        time = False
        ntime = 0
    else:
        ntime = fi.dimensions[timedim].size

    # collect variables with and without time dimension
    itvar = []     # w time dimension
    inottvar = []  # w/o time dimension
    if time or (time is None):
        for ivar in fi.variables.values():
            # remove variable if in removevar
            if ivar.name not in removevar:
                if timedim in ivar.dimensions:
                    itvar.append(ivar)

    if (not time) or (time is None):
        for ivar in fi.variables.values():
            # remove variable if in removevar
            if ivar.name not in removevar:
                if timedim not in ivar.dimensions:
                    inottvar.append(ivar)

    # copy static variables (not time-dependent)
    for ivar in inottvar:
        oname = ivar.name
        if ivar.name in renamevar.keys():
            oname = renamevar[ivar.name]
        ovar = fo.variables[oname]
        ovar[:] = ivar[:]

    # copy dynamic variables (time-dependent)
    for tt in range(ntime):
        for ivar in itvar:
            oname = ivar.name
            if ivar.name in renamevar.keys():
                oname = renamevar[ivar.name]
            ovar = fo.variables[oname]
            if ivar.ndim == 1:
                ovar[tt] = ivar[tt]
            else:
                ovar[tt, ...] = ivar[tt, ...]
    return


def create_new_variable(invardef, fo, izip=False, fill=None,
                        chunksizes=True):
    """
    Create variable in output file from dictionary with variable attributes.

    Parameters
    ----------
    invardef : dict
        Dictionary with name and dtype plus further attributes
        used in netCDF4.Dataset.createVariable; all other entries are set as
        variable attributes:
        'dimensions', 'zlib', 'complevel', 'shuffle', 'fletcher32',
        'contiguous', 'chunksizes', 'endian', 'least_significant_digit',
        'fill_value', 'chunk_cache'
    fo : file_handle
        File handle of opened netcdf output file
    izip : bool, optional
        True: the data will be compressed in the netCDF file using gzip
        compression independent of 'zlib' entry in input dictionary *invardef*
        (default: False).
    fill : float, bool or None, optional
        Determine the behaviour if variable has no _FillValue or missing_value.
        If None or False: no _FillValue will be set.
        If True: _FillValue will be set to default value of the Python netCDF4
        package for this type.
        If number: _FillValue will be set to number.
    chunksizes : bool, optional
        True: include possible chunksizes in output file (default).
        False: do not include chunksize information from input file in output
        file, even if given in input dictionary *invardef*.

    Returns
    -------
    variable handle
        Handle to newly created variable in output file.

    Examples
    --------
    .. code-block:: python

       nvar = {'name': 'new_field',
               'dtype': np.dtype(np.float),
               'dimensions': ('time', 'y', 'x'),
               'units': 'kg/m2/s',
               }
       ovar = create_new_variable(nvar, fo, fill=True, izip=True)

    """
    assert 'name' in invardef, 'name not in input dictionary'
    assert 'dtype' in invardef, 'dtype not in input dictionary'
    varname  = invardef.pop("name")
    datatype = invardef.pop("dtype")
    nckwargs = ['dimensions', 'zlib', 'complevel', 'shuffle', 'fletcher32',
                'contiguous', 'chunksizes', 'endian',
                'least_significant_digit', 'fill_value', 'chunk_cache']
    ncdict = dict()
    for nn in nckwargs:
        if nn in invardef:
            ncdict.update({nn: invardef.pop(nn)})
    if izip:
        ncdict.update({'zlib': True})
    if not chunksizes:
        if 'chunksizes' in ncdict:
            _ = ncdict.pop('chunksizes')
    if 'fill_value' not in ncdict:
        ncdict.update({'fill_value': None})
    # set _FillValue if None
    if ncdict['fill_value'] is None:
        if fill:
            if isinstance(fill, bool):
                ncdict['fill_value'] = get_fill_value_for_dtype(datatype)
            else:
                ncdict['fill_value'] = fill
    # create variable
    ovar = fo.createVariable(varname, datatype, **ncdict)
    # all other dict entries are attributes
    for k in invardef:
        ovar.setncattr(k, invardef[k])
    return ovar


def create_variables(fi, fo, time=None, timedim='time', izip=False, fill=None,
                     chunksizes=True, removevar=[], renamevar={}, removedim=[],
                     renamedim={}, replacedim={}):
    """
    Create variables in output from variables in input file.

    Parameters
    ----------
    fi : file_handle
        File handle of opened netcdf input file
    fo : file_handle
        File handle of opened netcdf output file
    time : None or bool, optional
        None:  create all variables (default).
        True:  create only variables having dimension *timedim*.
        False: create only variables that do not have dimension *timedim*.
    timedim : str, optional
        Name of time dimension (default: 'time').
    izip : bool, optional
        True: the data will be compressed in the netCDF file using gzip
        compression (default: False).
    fill : float, bool or None, optional
        Determine the behaviour if variable have no _FillValue or
        missing_value.
        If None or False: no _FillValue will be set.
        If True: _FillValue will be set to default value of the Python
        package netCDF4 for this type.
        If number: _FillValue will be set to number.
    chunksizes : bool, optional
        True: include possible chunksizes in output file (default).
        False: do not include chunksize information from input file in output
        file.
        Set to False, for example, if dimension size gets changed because the
        chunksize on a dimension can not be greater than the dimension size.
    removevar : list of str, optional
        Do not create variables given in *removevar* in output file.
    renamevar : dict, optional
        Rename variables in output file compared to input file.
        Variable names in input file are given as dictionary keys,
        corresponding variable names of output file are give as dictionary
        values.
    removedim : list of str, optional
        Remove dimensions from variable definitions in output file.
    renamedim : dict, optional
        Rename dimensions for variables in output file.
        Dimension names in input file are given as dictionary keys,
        corresponding dimension names of output file are give as dictionary
        values.
    replacedim : dict, optional
        Replace dimensions for variables in output file.
        Dimension names in input file are given as dictionary keys,
        corresponding dimension names of output file are given as dictionary
        values.
        The output names can be tuples or lists to extend dimensions of a
        variable.

    Returns
    -------
    nothing
        The output file will have the altered or unaltered variables
        of the input file defined.

    Examples
    --------
    .. code-block:: python

       create_variable(fi, fo, fill=True, izip=True, removedim=['patch'],
                       renamevar={'lon': 'longitude'},
                       replacedim={'land': ('y', 'x')})

    """
    removevar = _tolist(removevar)
    removedim = _tolist(removedim)
    for ivar in fi.variables.values():
        # remove variable if in removevar
        if ivar.name not in removevar:
            if time is None:
                itime = True
            else:
                if time:
                    itime = timedim in ivar.dimensions
                else:
                    itime = timedim not in ivar.dimensions
            if itime:
                invardef = get_variable_definition(ivar)
                if izip:
                    invardef.update({'zlib': True})
                # rename variable if in renamevar
                if ivar.name in renamevar.keys():
                    invardef['name'] = renamevar[ivar.name]
                # remove dimension if in removedim
                dims   = invardef['dimensions']
                chunks = invardef['chunksizes']
                for dd in removedim:
                    if dd in dims:
                        ip = dims.index(dd)
                        dims.remove(dd)
                        if chunks:
                            _ = chunks.pop(ip)
                # rename dimension if in renamedim
                for dd in renamedim.keys():
                    if dd in dims:
                        dims[dims.index(dd)] = renamedim[dd]
                # replace dimensions if in replacedim
                for dd in replacedim.keys():
                    if dd in dims:
                        rdim = _tolist(replacedim[dd])
                        ip = dims.index(dd)
                        dims = dims[:ip] + rdim + dims[ip + 1:]
                        if chunks:
                            rchunk = []
                            for cc in rdim:
                                # use fo not fi because new dims perhaps
                                # not yet in fi
                                rchunk.append(len(fo.dimensions[cc]))
                            chunks = chunks[:ip] + rchunk + chunks[ip+1:]
                invardef['dimensions'] = dims
                invardef['chunksizes'] = chunks
                # set _FillValue if None
                if invardef['fill_value'] is None:
                    if fill:
                        if isinstance(fill, bool):
                            invardef['fill_value'] = get_fill_value_for_dtype(
                                invardef['dtype'])
                        else:
                            invardef['fill_value'] = fill
                # exclude chunksizes
                if not chunksizes:
                    _ = invardef.pop('chunksizes')
                oname = invardef.pop("name")
                otype = invardef.pop("dtype")
                ovar = fo.createVariable(oname, otype, **invardef)
                for k in ivar.ncattrs():
                    iattr = ivar.getncattr(k)
                    if k != '_FillValue':
                        ovar.setncattr(k, iattr)
    return


def get_fill_value_for_dtype(dtype):
    """
    Get default _FillValue of netCDF4 for the given data type.

    Parameters
    ----------
    dtype : np.dtype
        numpy data type

    Returns
    -------
    default _FillValue of given numpy data type

    Examples
    --------
    .. code-block:: python

       fill_value = get_fill_value_for_dtype(var.dtype)

    """
    if dtype == np.dtype(np.int8):
        return nc.default_fillvals['i1']
    elif dtype == np.dtype(np.int16):
        return nc.default_fillvals['i2']
    elif dtype == np.dtype(np.int32):
        return nc.default_fillvals['i4']
    elif dtype == np.dtype(np.int64):
        return nc.default_fillvals['i8']
    elif dtype == np.dtype(np.uint8):
        return nc.default_fillvals['u1']
    elif dtype == np.dtype(np.uint16):
        return nc.default_fillvals['u2']
    elif dtype == np.dtype(np.uint32):
        return nc.default_fillvals['u4']
    elif dtype == np.dtype(np.uint64):
        return nc.default_fillvals['u8']
    elif dtype == np.dtype(np.float32):
        return nc.default_fillvals['f4']
    elif dtype == np.dtype(np.float64):
        return nc.default_fillvals['f8']
    else:
        import warnings
        warnings.warn("data type unknown: "+str(dtype))
        return None


def set_output_filename(ifile, ext):
    """
    Create output file name from input file name by adding *ext* before
    the file suffix.

    Parameters
    ----------
    ifile : str
        input file name
    ext : str
        string to add before file suffix

    Returns
    -------
    str
        output filename with ext before file suffix

    Examples
    --------
    >>> set_output_filename('in.nc', '-no_patch')
    in-no_patch.nc
    >>> set_output_filename('in.nc', '.nop')
    in.nop.nc

    """
    sifile = ifile.split('.')
    sifile[-2] = sifile[-2]+ext
    ofile = '.'.join(sifile)
    return ofile


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
import numpy as np
import scipy.optimize as opt
import pyjams.functions as functions


def line_dev_mask(x, y, z, p_guess=[1.,0.], plot=False):

    """
        Checks an np.ma.array y for outliers deviating from a straight line
        fitted to y(x). Outliers are detected when they deviate z*std(y) from
        the fitted line. Detection loops the fitting and std calculation until
        no outlier is detected anymore.

        Definition
        ----------
        def line_dev_mask(x, y, z, p_guess=[1.,0.], plot=False):


        Input
        -----
        x            np.ma.array(N)
        y            np.ma.array(N) or unmasked array of size N
        z            int: outlier threshold (line fit +- z*std(y))


        Optional Input
        --------------
        p_guess      list(2): initial parameters for line fit [slope, offset]
        plot         bool: if True, iterative plotting detection, if False, no
                           plotting


        Output
        ------
        new_mask     mask of y where additionally to the originally masked
                     elements, all elements are masked that are detected as
                     outliers deviating from the line fit by z*std(y)


        Examples
        --------
        >>> # Create some data
        >>> x = np.ma.array([1.,2.,3.,4.,5.,6.,7.,8.,9.,10.])
        >>> y = np.ma.array([1.,1.5,2.,2.5,3.,100.,8.,4.5,5.,5.5], mask=[0,1,0,0,0,0,0,0,0,0])

        >>> # detect outliers
        >>> new_mask = line_dev_mask(x,y,0.5,plot=False)
        >>> # apply new mask to x
        >>> print(np.ma.array(y.data, mask=new_mask))
        [1.0 -- 2.0 2.5 3.0 -- -- 4.5 5.0 5.5]


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Arndt Piayda, Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  AP, Feb 2014
        Modified, MC, Feb 2014 - use jams.functions
    """

    if plot:
        import matplotlib.pyplot as plt

    # create working array
    y_new = np.ma.copy(y)

    # loop until no outliers are detected anymore
    go_on = True
    while go_on:
        # fit line to data
        p_opt = opt.fmin(functions.cost_abs, p_guess, args=(functions.line_p,x,y_new), disp=False)

        # calculate maximum and minimum deviation limit depending on std
        max_dev = functions.line_p(x,[p_opt[0]+np.ma.std(y_new)*z,p_opt[1]])
        min_dev = functions.line_p(x,[p_opt[0]-np.ma.std(y_new)*z,p_opt[1]])

        if plot:
            fig = plt.figure('line_dev_mask')
            sub = fig.add_subplot(111)
            sub.plot(x,y_new,'ko')
            sub.plot(x,functions.line_p(x,p_opt),'k-')
            sub.plot(x,min_dev,'r-')
            sub.plot(x,max_dev,'r-')
            plt.show()

        # check for limit trespass
        new_mask = (y_new>max_dev) | (y_new<min_dev)
        y_new.mask += new_mask

        # go on when outliers are detected
        go_on = True if np.sum(new_mask)>0 else False

    # return mask where outliers are masked
    return y_new.mask

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
"""
This is the unittest for kernel_regression module.

python -m unittest -v tests/test_kernel_regression.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_kernel_regression.py

"""
import unittest


def _flatten(itr):
    import numpy as np
    fitr = np.array(itr).flatten()
    if len(fitr) == 0:
        return list(fitr)
    else:
        if isinstance(fitr[0], str):
            return [ i for i in fitr ]
        else:
            return [ i if np.isfinite(i) else np.finfo(float).max
                     for i in fitr ]


class TestKernelRegression(unittest.TestCase):
    """
    Tests for kernel_regression.py
    """

    def setUp(self):
        import numpy as np
        # seed for reproducible results
        self.seed = 1234
        np.random.seed(seed=self.seed)

    def test_kernel_regression_h(self):
        import numpy as np
        import pandas as pd
        from pyjams import kernel_regression_h

        # 1971 Canadian Census Public Use Tapes
        # as in Wikipedia article about kernel regression,
        # taken from R package crs:
        # https://github.com/JeffreyRacine/R-Package-crs
        infile = 'tests/cps71.csv'
        dat = np.loadtxt(infile, skiprows=1, delimiter=',')
        y = dat[:, 1]
        x = dat[:, 2]
        df = pd.DataFrame({'x': x, 'y': y})

        # bootstrap h
        hout  = kernel_regression_h(x, y)
        hsoll = 4.46598
        assert isinstance(hout, np.float64)
        self.assertEqual(np.around(hout, 5), hsoll)

        hout  = kernel_regression_h(df['x'], df['y'])
        hsoll = 4.46598
        assert isinstance(hout, np.float64)
        self.assertEqual(np.around(hout, 5), hsoll)

        # silverman
        hout  = kernel_regression_h(x, y, silverman=True)
        hsoll = 4.46598
        assert isinstance(hout, np.float64)
        self.assertEqual(np.around(hout, 5), hsoll)

        hout  = kernel_regression_h(df['x'], df['y'], silverman=True)
        hsoll = 4.46598
        assert isinstance(hout, np.float64)
        self.assertEqual(np.around(hout, 5), hsoll)

        #

        # made up multidimensional data
        n = 10
        x = np.zeros((n, 2))
        x[:, 0] = np.arange(n, dtype=float) / float(n - 1)
        x[:, 1] = 1. / (np.arange(n, dtype=float) / float(n - 1) + 0.1)
        y = 1. + x[:, 0]**2 - np.sin(x[:, 1])**2
        df = pd.DataFrame({'x1': x[:, 0], 'x2': x[:, 1]})

        # cross-validation
        hout  = kernel_regression_h(x, y)
        hsoll = [0.17268, 9.51691]
        assert isinstance(hout, np.ndarray)
        self.assertEqual(list(np.around(hout, 5)), hsoll)

        # silverman
        hout = kernel_regression_h(x, y, silverman=True)
        hsoll = [0.22919, 1.90338]
        assert isinstance(hout, np.ndarray)
        self.assertEqual(list(np.around(hout, 5)), hsoll)

        hout = kernel_regression_h(df, y, silverman=True)
        hsoll = [0.22919, 1.90338]
        assert isinstance(hout, np.ndarray)
        self.assertEqual(list(np.around(hout, 5)), hsoll)

        # errors
        # shapes
        self.assertRaises(AssertionError, kernel_regression_h,
                          x, y[1:])

    def test_kernel_regression(self):
        import numpy as np
        import pandas as pd
        from pyjams import kernel_regression_h, kernel_regression

        # 1971 Canadian Census Public Use Tapes
        # as in Wikipedia article about kernel regression,
        # taken from R package crs:
        # https://github.com/JeffreyRacine/R-Package-crs
        infile = 'tests/cps71.csv'
        dat = np.loadtxt(infile, skiprows=1, delimiter=',')
        y = dat[:, 1]
        x = dat[:, 2]
        df = pd.DataFrame({'x': x, 'y': y})

        nout = 5
        xout = (x.min() + (x.max() - x.min()) *
                np.arange(nout, dtype=float) / float(nout - 1))

        # using kernel_regression_h
        h = kernel_regression_h(x, y)
        fout = kernel_regression(x, y, h)
        fsoll = [13.0172, 13.3331, 13.693, 13.6816, 13.3306]
        assert isinstance(fout, np.ndarray)
        self.assertEqual(list(np.around(fout[::50], 4)), fsoll)

        h = kernel_regression_h(df['x'], df['y'])
        fout = kernel_regression(df['x'], df['y'], h)
        fsoll = [13.0172, 13.3331, 13.693, 13.6816, 13.3306]
        assert isinstance(fout, pd.Series)
        self.assertEqual(list(np.around(fout[::50], 4)), fsoll)

        # w/o kernel_regression_h
        fout = kernel_regression(x, y)
        fsoll = [13.0172, 13.3331, 13.693, 13.6816, 13.3306]
        assert isinstance(fout, np.ndarray)
        self.assertEqual(list(np.around(fout[::50], 4)), fsoll)

        # silverman
        fout = kernel_regression(x, y, silverman=True)
        fsoll = [13.0172, 13.3331, 13.693, 13.6816, 13.3306]
        assert isinstance(fout, np.ndarray)
        self.assertEqual(list(np.around(fout[::50], 4)), fsoll)

        # xout
        fout = kernel_regression(x, y, silverman=True, xout=xout)
        fsoll = [13.0172, 13.6, 13.66, 13.6791, 13.2663]
        assert isinstance(fout, np.ndarray)
        self.assertEqual(list(np.around(fout, 4)), fsoll)

        # different input/output types
        fout = kernel_regression(x, list(y), silverman=True, xout=xout)
        fsoll = [13.0172, 13.6, 13.66, 13.6791, 13.2663]
        assert isinstance(fout, list)
        self.assertEqual(list(np.around(fout, 4)), fsoll)

        fout = kernel_regression(x, tuple(y), silverman=True, xout=xout)
        fsoll = [13.0172, 13.6, 13.66, 13.6791, 13.2663]
        assert isinstance(fout, tuple)
        self.assertEqual(list(np.around(fout, 4)), fsoll)

        fout = kernel_regression(x, np.ma.array(y), silverman=True, xout=xout)
        fsoll = [13.0172, 13.6, 13.66, 13.6791, 13.2663]
        assert isinstance(fout, np.ma.masked_array)
        self.assertEqual(list(np.around(fout, 4)), fsoll)

        fout = kernel_regression(list(x), y)
        fsoll = [13.0172, 13.3331, 13.693, 13.6816, 13.3306]
        assert isinstance(fout, np.ndarray)
        self.assertEqual(list(np.around(fout[::50], 4)), fsoll)

        fout = kernel_regression(tuple(x), y)
        fsoll = [13.0172, 13.3331, 13.693, 13.6816, 13.3306]
        assert isinstance(fout, np.ndarray)
        self.assertEqual(list(np.around(fout[::50], 4)), fsoll)

        fout = kernel_regression(np.ma.array(x), y)
        fsoll = [13.0172, 13.3331, 13.693, 13.6816, 13.3306]
        assert isinstance(fout, np.ndarray)
        self.assertEqual(list(np.around(fout[::50], 4)), fsoll)

        fout = kernel_regression(df['x'], df['y'])
        fsoll = [13.0172, 13.3331, 13.693, 13.6816, 13.3306]
        assert isinstance(fout, pd.Series)
        self.assertEqual(list(np.around(fout[::50], 4)), fsoll)

        #

        # made up multidimensional data
        n = 10
        x = np.zeros((n, 2))
        x[:, 0] = np.arange(n, dtype=float) / float(n - 1)
        x[:, 1] = 1. / (np.arange(n, dtype=float) / float(n - 1) + 0.1)
        y = 1. + x[:, 0]**2 - np.sin(x[:, 1])**2
        df = pd.DataFrame({'x1': x[:, 0], 'x2': x[:, 1]})

        nout = 5
        xout = np.empty((nout, 2))
        xout[:, 0] = (np.amin(x[:, 0]) +
                      (np.amax(x[:, 0]) - np.amin(x[:, 0])) *
                      np.arange(nout, dtype=float) / float(nout - 1))
        xout[:, 1] = (np.amin(x[:, 1]) +
                      (np.amax(x[:, 1]) - np.amin(x[:, 1])) *
                      np.arange(nout, dtype=float) / float(nout - 1))

        # using kernel_regression_h
        h = kernel_regression_h(x, y)
        fout = kernel_regression(x, y, h)
        fsoll = [0.5224, 0.5257, 0.5418, 0.5178,
                 0.4764, 0.4923, 0.6034, 0.7775,
                 0.9545, 1.0960]
        assert isinstance(fout, np.ndarray)
        self.assertEqual(list(np.around(fout, 4)), fsoll)

        h = kernel_regression_h(df, y)
        fout = kernel_regression(df, y, h)
        fsoll = [0.5224, 0.5257, 0.5418, 0.5178,
                 0.4764, 0.4923, 0.6034, 0.7775,
                 0.9545, 1.0960]
        assert isinstance(fout, np.ndarray)
        self.assertEqual(list(np.around(fout, 4)), fsoll)

        # using given kernel
        fout = kernel_regression(x, y, 0.2)
        fsoll = [0.704, 0.0129, 1.0478, 0.5477,
                 0.2902, 0.3825, 0.5992, 0.843,
                 1.0402, 1.1707]
        assert isinstance(fout, np.ndarray)
        self.assertEqual(list(np.around(fout, 4)), fsoll)

        # w/o kernel_regression_h
        fout = kernel_regression(x, y)
        fsoll = [0.5224, 0.5257, 0.5418, 0.5178,
                 0.4764, 0.4923, 0.6034, 0.7775,
                 0.9545, 1.0960]
        assert isinstance(fout, np.ndarray)
        self.assertEqual(list(np.around(fout, 4)), fsoll)

        # silverman
        fout = kernel_regression(x, y, silverman=True)
        fsoll = [0.6912, 0.4228, 0.5458, 0.5343,
                 0.5215, 0.5554, 0.6421, 0.7619,
                 0.8878, 1.0001]
        assert isinstance(fout, np.ndarray)
        self.assertEqual(list(np.around(fout, 4)), fsoll)

        # xout
        fout = kernel_regression(x, y, silverman=True, xout=xout)
        fsoll = [0.6055, 0.5428, 0.4945, 0.5267, 0.6951]
        assert isinstance(fout, np.ndarray)
        self.assertEqual(list(np.around(fout, 4)), fsoll)

        # different input/output types
        fout = kernel_regression(x, list(y))
        fsoll = [0.5224, 0.5257, 0.5418, 0.5178,
                 0.4764, 0.4923, 0.6034, 0.7775,
                 0.9545, 1.0960]
        assert isinstance(fout, list)
        self.assertEqual(list(np.around(fout, 4)), fsoll)

        fout = kernel_regression(x, tuple(y))
        fsoll = [0.5224, 0.5257, 0.5418, 0.5178,
                 0.4764, 0.4923, 0.6034, 0.7775,
                 0.9545, 1.0960]
        assert isinstance(fout, tuple)
        self.assertEqual(list(np.around(fout, 4)), fsoll)

        fout = kernel_regression(x, np.ma.array(y))
        fsoll = [0.5224, 0.5257, 0.5418, 0.5178,
                 0.4764, 0.4923, 0.6034, 0.7775,
                 0.9545, 1.0960]
        assert isinstance(fout, np.ma.masked_array)
        self.assertEqual(list(np.around(fout, 4)), fsoll)

        fout = kernel_regression(np.ma.array(x), y)
        fsoll = [0.5224, 0.5257, 0.5418, 0.5178,
                 0.4764, 0.4923, 0.6034, 0.7775,
                 0.9545, 1.0960]
        assert isinstance(fout, np.ndarray)
        self.assertEqual(list(np.around(fout, 4)), fsoll)

        fout = kernel_regression(df, y)
        fsoll = [0.5224, 0.5257, 0.5418, 0.5178,
                 0.4764, 0.4923, 0.6034, 0.7775,
                 0.9545, 1.0960]
        assert isinstance(fout, np.ndarray)
        self.assertEqual(list(np.around(fout, 4)), fsoll)

        # errors
        # shapes do not match
        self.assertRaises(AssertionError, kernel_regression, x, y[1:])
        # h has wrong size
        self.assertRaises(AssertionError, kernel_regression, x, y,
                          [0.1, 0.2, 0.3])
        # xout dimension does not match
        xxout = np.empty((nout, 3))
        xxout[:, 0:2] = xout
        xxout[:, 2] = xout[:, 1]
        self.assertRaises(AssertionError, kernel_regression, x, y,
                          [0.1, 0.2, 0.3], xout=xxout)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
Author
------
David Schaefer

Purpose
-------
A sanitizing layer for the netCDF4 library. Adds a number of convenince methods
and aims for a cleaner user interface. All classes available are children of their
associated netCDF4 counterparts.
"""
import uuid
from netCDF4 import Dataset, Group, Dimension, Variable, date2num, num2date
from collections import OrderedDict

def _tupelize(arg):
    if isinstance(arg,str):
        return (arg,)
    try:
        return tuple(arg)
    except TypeError:
        return (arg,)


def copyGroup(ncin, group, skipdims=None, skipgroups=None, skipvars=None,
              skipattrs=None, fixdims=False, vardata=False, varparams=None):
    """
    Arguments
    ---------
    ncin                  : Instance of an object with a createGroup method
                            (i.e. NcDataset, NcGroup)
    group                 : Instance of an object with dimensions/variables/attributes/groups attributes
                            (i.e. NcDataset, NcGroup)
    skipdims (optional)   : string or list/tuple of strings
                            Name(s) of dimension(s) to skip
    skipgroups (optional) : string or list/tuple of strings
                            Name(s) of group(s) to skip
    skipvars (optinal)    : string or list/tuple of strings
                            Name(s) of variable(s) to skip
    skipattrs (optinal)   : string or list/tuple of strings
                            Name(s) of attribute(s) to skip
    fixdims (optional)    : convert unlimited dimensions to fixed size dimensions
    vardata (optional)    : boolean, copy variable data
    varparams(optional)   : dict, variable paramaters will be passed to
                            createVariable (i.e. zlib, complevel, chunksizes, ...)

    Return
    ------
    NcGroup

    Purpose
    -------
    Copy the given group to ncin
    """
    out = ncin.createGroup(group.name)
    out.set_fill_off()
    out.copyDimensions(group.dimensions, skip=skipdims, fix=fixdims)
    out.copyVariables(
        group.variables, skip=skipvars, data=vardata, varparams=varparams
    )
    out.copyAttributes(group.attributes, skipattrs)
    out.copyGroups(group.groups, skipgroups)
    return out


def copyDataset(ncin, group, skipdims=None, skipgroups=None, skipvars=None,
                skipattrs=None, fixdims=False, vardata=False, varparams=None):
    """
    Arguments
    ---------
    ncin                  : Instance of an object with a createGroup method
                            (i.e. NcDataset, NcGroup)
    group                 : Instance of an object with dimensions/variables/attributes/groups attributes
                            (i.e. NcDataset, NcGroup)
    skipdims (optional)   : string or list/tuple of strings
                            Name(s) of dimension(s) to skip
    skipgroups (optional) : string or list/tuple of strings
                            Name(s) of group(s) to skip
    skipvars (optinal)    : string or list/tuple of strings
                            Name(s) of variable(s) to skip
    skipattrs (optinal)   : string or list/tuple of strings
                            Name(s) of attribute(s) to skip
    fixdims (optional)    : convert unlimited dimensions to fixed size dimensions
    vardata (optional)    : boolean, copy variable data
    varparams(optional)   : dict, variable paramaters will be passed to
                            createVariable (i.e. zlib, complevel, chunksizes, ...)


    Return
    ------
    NcDataset/NcGroup

    Purpose
    -------
    Copy the content of given group to ncin
    """
    ncin.set_fill_off()
    ncin.copyDimensions(group.dimensions, skip=skipdims, fix=fixdims)
    ncin.copyVariables(
        group.variables, skip=skipvars, data=vardata, varparams=varparams
    )
    ncin.copyAttributes(group.attributes, skipattrs)
    ncin.copyGroups(group.groups, skipgroups)
    return ncin

def copyGroups(ncin, groups, skip=None):
    """
    Arguments
    ---------
    ncin            : Instance of an object with a createGroup method
                      (i.e. NcDataset, NcGroup)
    groups          : Dictionary
                      key   : group name (string)
                      value : instance of an object with dimensions/variables/attributes/groups attributes
    skip (optional) : string or list/tuple of strings
                      Name(s) of group(s) to skip
    
    Return
    ------
    None

    Purpose
    -------
    Copy the given groups to ncin
    """
    for g in groups.values():
        if g.name not in _tupelize(skip):
            ncin.copyGroup(g)


def copyDimension(ncin, dim, fix=False, fail=True):
    """
    Arguments
    ---------
    ncin            : Instance of an object with a createDimension method
                      (i.e. NcDataset, NcGroup)
    group           : Instance of NcDimension
    fix (optional)  : convert an unlimited dimension to a fixed size dimension
    fail (Optional[bool]): raise an exception if dimension already exists

    Return
    ------
    netCDF4.Dimension

    Purpose
    -------
    Copy the given dimension to ncin
    """
    length = None if dim.isunlimited() and not fix else len(dim)
    try:
        return ncin.createDimension(dim.name, length)
    except Exception:
        if fail:
            raise
        return ncin.dimensions[dim.name]


def copyDimensions(ncin, dimensions, skip=None, fix=False, fail=True):
    """
    Arguments
    ---------
    ncin            : Instance of an object with a createDimension method
                      (i.e. NcDataset, NcGroup)
    dimension       : Dictionary
                      key   : dimension name (string)
                      value : instance of NcDimension
    skip (optional) : string or list/tuple of strings
                      Name(s) of dimension(s) to skip
    fix (optional)  : convert unlimited dimensions to fixed size dimensions
    fail (Optional[bool]): raise exception if dimension already exists

    Return
    ------
    None

    Purpose
    -------
    Copy the given dimensions to ncin
    """
    for d in dimensions.values():
        if d.name not in _tupelize(skip):
            ncin.copyDimension(d, fix=fix, fail=fail)


def copyAttributes(ncin, attributes, skip=None):
    """
    Arguments
    ---------
    ncin              : Instance of an object with a createAttribute method
                        (i.e. NcDataset, NcGroup, NcVariable)
    attributes        : Dictionary
                        key   : string
                        value : string/any numeric type
    skip (optional)   : string or list/tuple of strings
                        Name(s) of attribute(s) to skip

    Return
    ------
    None

    Purpose
    -------
    Copy the given attributes to ncin
    """
    for k, v in attributes.items():
        if k not in _tupelize(skip):
            if k == "missing_value":
                try:
                    v = ncin.dtype.type(v)
                except Exception:
                    pass
            ncin.createAttribute(k, v)


def copyVariable(ncin, var, data=True, dims=False, fail=True, **kwargs):
    """
    Arguments
    ---------
    ncin            : Instance of an object with a createCopy method
                      (i.e. NcDataset, NcGroup, NcVariable)
    var             : Instance of NcVariable
    data (optional) : boolean, copy variable data
    dims (Optional[bool]): copy missing dimensions
    fail (Optional[bool]): raise an exception if variable exists, ignored at the moment
    kwargs          : will be passed to createVariable. Allows to set
                      parameters like chunksizes, deflate_level, ...
    
    Return
    ------
    NcVariable
    
    Purpose
    -------
    Copy the given variables to ncin. Copy the data if data=True
    """
    invardef = var.definition
    
    if data is not True:
        invardef["chunksizes"] = None
    invardef.update(kwargs)

    if dims:
        nc = var.parent
        try:
            shape = data.shape
        except AttributeError:
            shape = var.shape
        for name, length in zip(var.dimensions, shape):
            l = None if nc.dimensions[name].isunlimited() else length
            ncin.createDimension(name, l, fail=fail)

    vname = invardef.pop("name")
    try:
        invar = ncin.createVariable(vname, invardef.pop("dtype"), **invardef)
    except RuntimeError:
        if fail:
            raise
        invar = ncin.variables[vname]

    invar.copyAttributes(var.attributes)
    if data is True and var.shape:
        invar[:] = var[:]
    elif data is not False:
        # i.e. if an array is given
        invar[:] = data
    return invar


def copyVariables(ncin, variables, skip=None, data=True, dims=False, fail=True, varparams=None):
    """
    Arguments
    ---------
    ncin                    : Instance of an object with a createCopy method
                              (i.e. NcDataset, NcGroup, NcVariable)
    variables               : Dictionary
                              key   : variables name (string)
                              value : instance of NcVariable
    skip (optional)         : string or list/tuple of strings
                              Name(s) of variable(s) to skip
    data (optional)         : boolean
    dims (Optional[bool])   : copy nonexisting dimensions
    fail (Optional[bool])   : raise an exception if a variable already exists
    varparams(optional)     : dict, variable paramaters will be passed to
                              createVariable (i.e. zlib, complevel, chunksizes, ...)

    Return
    ------
    NcVariable
    
    Purpose
    -------
    Copy the given variables to ncin. Copy the data if data=True
    """

    if varparams is None:
        varparams = dict()
    for v in variables.values():
        if v.name not in _tupelize(skip):
            ncin.copyVariable(v, data, dims, fail, **varparams)


def createDimensions(ncin, dim_dict, fail=True):
    for name, length in dim_dict.items():
        ncin.createDimension(name, length, fail=fail)


        
        
def getVariableDefinition(ncvar):
    out = ncvar.filters() if ncvar.filters() else {}
    out.update({
        "name"       : ncvar.name,
        "dtype"      : ncvar.dtype,
        "dimensions" : ncvar.dimensions,
        "chunksizes" : ncvar.chunking() if not isinstance(ncvar.chunking(), str) else None,
        "fill_value" : getattr(ncvar, "_FillValue", None),
    })
    return out

def getDates(ncin, timesteps=None, timevar="time", units=None, calendar=None):
    """
    Arguments
    ---------
    ncin                 : Instance of an object holding variables (NcDataset/NcGroup)
    timesteps (optional) : list/tuple/nd.array of Numerical values.
                           The time_steps to return dates for. If not given the content of the
                           entire time variable will be returned.
    units (optional)     : string
                           time units following the CF Conventions. Needs to be given, if not
                           available as an attribute of the time variable.
    calendar (optional)  : string
                           calendar name following the CF conventions. Needs to be given, if not
                           available as an attribute of the time variable.

    Return
    ------
    List of datetime objects

    Purpose
    -------
    Return datetime objects associated to the time variable of ncin
    """
    var = ncin.variables[timevar]        
    if not units:
        try:
            units = var.units
        except AttributeError:
            raise AttributeError(
                "Time variable does not specify an units attribute! Pass as argument."
            )

    if not calendar:
        try:
            calendar = var.calendar
        except AttributeError:
            calendar = "standard"

    if not timesteps:
        timesteps = var[:]

    dates = num2date(timesteps,units,calendar)
    
    try:
        return [d.date() for d in dates]
    except AttributeError:
        return dates

def setFillValue(ncin, value):
    """
    Arguments
    ---------
    ncin  : Instance of an object with a _FillValue attribute
            (i.e. NcVariable)
    
    Return
    ------
    Numeric

    Purpose
    -------
    Return the value of the attribute _FillValue
    """
    ncin.setncattr("_FillValue", value)

def getFillValue(ncin):
    """
    Arguments
    ---------
    ncin  : Instance of an object with a _FillValue attribute
            (i.e. NcVariable)
    
    Return
    ------
    Numeric

    Purpose
    -------
    Return the value of the attribute _FillValue
    """
    try:
        return ncin.getncattr("_FillValue")
    except AttributeError:
        return None
        
def setAttribute(ncin, name, value):
    """
    Arguments
    ---------
    ncin  : Instance of an object with a setncatts method
            (i.e. NcDataset/NcGroup/NcVariable)
    name  : string
    value : string or any numeric type
    
    Return
    ------
    None

    Purpose
    -------
    Set/Write the attribute given as name, value
    """
    ncin.setncattr(name, value)

def setAttributes(ncin, attdict):
    """
    Arguments
    ---------
    ncin    : Instance of an object with a setncatts method
              (i.e. NcDataset/NcGroup/NcVariable)
    attdict : dictionary
              key: attribute name (string)
              value: attribute value (string or any numeric type)

    Return
    ------
    None

    Purpose
    -------
    Set/Write the attributes given in attdict
    """
    ncin.setncatts(attdict)
 
def filterVariables(ncin, dims=None, ndim=None):
    """
    Arguments
    ---------
    ncin            : Instance of an object with a variables attribute
                      (i.e. NcDataset/NcGroup)
    dims (optional) : tuple/list of dimension strings
    ndim (optional) : int number of dimensions

    Return
    ------
    OrderedDict:
        key   : variable name (string)
        value : NcVariable instance
    
    Purpose
    -------
    Return all Variables that are based on the dimension(s) given in dims
    and/or have ndims dimensions.
    """
    out = OrderedDict()
    dims = set(dims or {})
    
    for v in ncin.variables.values():
        if dims.issubset(set(v.dimensions)):
            if ndim:
                if ndim == len(v.dimensions):
                    out[v.name] = v
            else:
                out[v.name] = v
                
    return out

def filterDimensions(ncin, lengths):
    """
    Arguments
    ---------
    lengths : integer, tuple/list of integers

    Return
    ------
    OrderedDict:
        key   : dimension name (string)
        value : NcDimension instance

    Purpose
    -------
    Return all Dimensions with a length given in the argument lengths.
    """
    try:
        lengths[0]
    except TypeError:
        lengths = (lengths,)

    return OrderedDict(
        [(d.name, d) for d in ncin.dimensions.values() if len(d) in lengths]
    ) 

def getGroups(ncin):
    out = OrderedDict()
    for g in getattr(ncin, "groups").values():
        out[g.name] = NcGroup(ncin, g.name, id=g._grpid)
    return out

def getVariables(ncin):
    out = OrderedDict()
    for v in getattr(ncin, "variables").values():
        out[v.name] = NcVariable(ncin, v.name, v.dtype, v.dimensions, id=v._varid)
    return out


def getDimensions(ncin):
    
    out = OrderedDict()
    for dim in getattr(ncin, "dimensions").values():
        out[dim.name] = NcDimension(ncin, dim.name, id=dim._dimid)
    return out


def getAttributes(ncin):
    out = OrderedDict()
    for k in ncin.ncattrs():
        if not k.startswith("_"):
            out[k] = ncin.getncattr(k)
    return out


def getParent(ncin):
    return ncin._grp


def attributeSetter(ncin, name, value):
    ncin.__dict__[name] = value
  
def attributeGetter(ncin, name):
    try:
        return ncin.__dict__[name]
    except KeyError:
        try:
            return getattr(super(ncin.__class__, ncin), name)
        except KeyError:
            raise AttributeError(
                "'{:}' object has no attribute '{:}'".format(ncin.__class__, name)
            )

def createGroup(ncin, name):
    grp = NcGroup(ncin, name)
    ncin.groups[name] = grp
    return grp

def createVariable(ncin, *args, **kwargs):
    var = NcVariable(ncin, *args, **kwargs)
    ncin.variables[var.name] = var
    return var

def createDimension(ncin, name, length, fail=True):
    try:
        dim = NcDimension(ncin, name, length)
        ncin.dimensions[dim.name] = dim
    except Exception:
        if fail:
            raise
        dim = ncin.dimensions[name]
    return dim


class NcDataset(Dataset):
    def __init__(
            self,
            filename = None,
            mode     = "r",
            clobber  = True,
            diskless = False,
            persist  = False,
            weakref  = False,
            format   = "NETCDF4",
    ):

        if filename is None:
            # in memory dataset
            filename = str(uuid.uuid4())
            mode = "w"
            diskless = True

        super(NcDataset, self).__init__(
            filename = filename,
            mode     = mode,
            clobber  = clobber,
            diskless = diskless,
            persist  = persist,
            weakref  = weakref,
            format   = format,
        )

        self.fname = filename
        for k, v in zip(self.groups, getGroups(self).values()):
            self.groups[k] = v
        for k, v in zip(self.dimensions, getDimensions(self).values()):
            self.dimensions[k] = v
        for k, v in zip(self.variables, getVariables(self).values()):
            self.variables[k] = v

    def tofile(self, fname):
        # preserve dataset options
        with NcDataset(fname, "w") as out:
            out.copyDataset(self, vardata=True)

    # def __enter__(self):
    #     return self

    # def __exit__(self, *args, **kwargs):
    #     self.close()

    copyDataset      = copyDataset
    copyDimension    = copyDimension
    copyDimensions   = copyDimensions
    copyAttributes   = copyAttributes
    copyVariable     = copyVariable
    copyVariables    = copyVariables
    copyGroup        = copyGroup
    copyGroups       = copyGroups
    createAttribute  = setAttribute
    createAttributes = setAttributes
    createDimensions = createDimensions
    createVariable   = createVariable
    createGroup      = createGroup
    createDimension  = createDimension
    filterVariables  = filterVariables
    filterDimensions = filterDimensions
    getDates         = getDates
    attributes       = property(fget=getAttributes)
    # restore a "normal" attribute access behaviour
    __setattr__      = attributeSetter
    __getattr__      = attributeGetter


class NcGroup(Group):
    def __init__(self, *args, **kwargs):
        super(NcGroup,self).__init__(*args, **kwargs)
        for k, v in zip(self.groups, getGroups(self).values()):
            self.groups[k] = v
        for k, v in zip(self.dimensions, getDimensions(self).values()):
            self.dimensions[k] = v
        for k,v in zip(self.variables, getVariables(self).values()):
            self.variables[k] = v

    copyDimension    = copyDimension
    copyDimensions   = copyDimensions
    copyAttributes   = copyAttributes
    copyVariable     = copyVariable
    copyVariables    = copyVariables
    copyGroup        = copyGroup
    copyGroups       = copyGroups
    createAttribute  = setAttribute
    createAttributes = setAttributes
    createDimensions = createDimensions
    createVariable   = createVariable
    createGroup      = createGroup
    createDimension  = createDimension
    filterVariables  = filterVariables
    filterDimensions = filterDimensions
    getDates         = getDates
    attributes       = property(fget=getAttributes)
    parent           = property(fget=getParent)
    # restore a "normal" attribute access behaviour
    __setattr__      = attributeSetter
    __getattr__      = attributeGetter


class NcVariable(Variable):
    def __init__(self, *args, **kwargs):
        super(NcVariable, self).__init__(*args, **kwargs)

    copyAttributes   = copyAttributes
    createAttribute  = setAttribute
    createAttributes = setAttributes
    attributes       = property(fget=getAttributes)
    definition       = property(fget=getVariableDefinition)
    fill_value       = property(fget=getFillValue, fset=setFillValue)
    parent           = property(fget=getParent)
    # restore a "normal" attribute access behaviour
    __setattr__      = attributeSetter
    __getattr__      = attributeGetter


# Just to be consistent...
class NcDimension(Dimension):
    def __init__(self, *args, **kwargs):
        super(NcDimension, self).__init__(*args, **kwargs)
    
    parent = property(fget=getParent)



#!/usr/bin/env python
"""
    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT License.

    Copyright (c) 2016-2017 Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

    Modified, MC, Dec 2017 - multinormal
"""
import numpy as np


__all__ = ['exponential', 'laplace',
           'gauss', 'norm', 'normal',                           # Normal (Gauss)
           'multigauss', 'multinorm', 'multinormal',            # Multivariate Normal (Gauss)
           'ep', 'sep', 'sep_fs', 'sep_fs_mean', 'sep_fs_std',  # (Skew) Exponential Power
           'st', 'st_fs', 'st_fs_mean', 'st_fs_std', 't']       # (Skew) Student-t


# --------------------------------------------------------------------


def ep(x, loc=0., sca=1., beta=0., sig=None):
    """
        The exponential power distribution with given location, scale, and kurtosis.


        Definition
        ----------
        def ep(x, loc=0., sca=1., beta=0., sig=None):


        Input
        -----
        x          array_like quantiles


        Optional Input
        --------------
        loc        location
        sca        scale
        beta       kurtosis parameter
        sig        standard deviation, overwrites scale


        Output
        ------
        Exponential power pdf at x


        Examples
        --------
        >>> print(np.allclose(sep(1., 2., 2., 0.5, 2.), sep((1.-2.)/2., xi=0.5, beta=2.)/2.))
        True

        >>> print(np.allclose(sep(1.3, 0., 1., 1., 0.), normal(1.3, 0., 1.)))
        True

        >>> print(np.allclose(sep(1.3, 0., 1., 1., 1.), laplace(1.3, 0., 1./np.sqrt(2.))))
        True

        >>> print(np.allclose(sep(1.3, 0., np.sqrt(2.), 1., 1.), laplace(1.3, 0., 1.)))
        True


        History
        -------
        Written,  MC, May 2016
    """
    if sig is not None: sca = sig
    return ep01((x-loc)/sca, beta)/sca


def ep01(x, beta=0.):
    """
        The exponential power distribution with given kurtosis parameter, location zero and unit scale.


        Definition
        ----------
        def ep01(x, beta=0.):


        Input
        -----
        x          array_like quantiles


        Optional Input
        --------------
        beta       kurtosis parameter


        Output
        ------
        Exponential power pdf with loc=0, sca=1 at x


        History
        -------
        Written,  MC, May 2016
    """
    import scipy.special as ss

    if beta != -1.0:
        b1 = 0.5*(1.0 + beta)
        b3 = 1.5*(1.0 + beta)
        g1 = ss.gamma(b1)
        g3 = ss.gamma(b3)
        # -> 0
        c_beta = (g3/g1)**(1.0/(1.0+beta))
        # -> sqrt(1/12)
        om_beta = np.sqrt(g3)/((1.0+beta)*np.sqrt(g1**3))
    else:
        c_beta  = 0.0
        om_beta = np.sqrt(1.0/12.0)

    # pdf
    if np.abs(beta+1.0) < 0.003: # 2/(1-0.997) ~ 666
        # Uniform between [-x1,x1]
        height = om_beta
        x1 = 0.5/height # int(pdf) = 1 = 2*x1*height
        pdf = np.where((x > (-x1)) & (x < x1), height, 0.0)
        if not np.iterable(x): pdf = pdf[0]
        return pdf
    else:
        return om_beta * np.exp(-c_beta*np.abs(x)**(2.0/(1.0+beta)))


# --------------------------------------------------------------------


def exponential(x, loc=0., sca=1., sig=None):
    """
        Exponential probability density function (pdf).


        Definition
        ----------
        def exponential(x, loc=0., sca=1., sig=None):


        Input
        -----
        x          array_like quantiles


        Optional Input
        --------------
        loc        location
        sca        scale
        sig        standard deviation, overwrites scale


        Output
        ------
        Exponential pdf at x


        History
        -------
        Written,  MC, May 2016
    """

    if sig is not None: sca = sig
    return exponential01((x-loc)/sca)/sca


def exponential01(x):
    """
        Exponential probability density function (pdf) at location zero and with unit scale.


        Definition
        ----------
        def exponential01(x):


        Input
        -----
        x          array_like quantiles


        Output
        ------
        Exponential pdf with loc=0, sca=1. at x


        History
        -------
        Written,  MC, May 2016
    """

    iexp = np.where(x<0.0, 0., np.exp(-x))
    if np.iterable(x):
        return iexp
    else:
        return float(iexp)


# --------------------------------------------------------------------


def gauss(*args, **kwargs):
    """
        Wrapper for normal

        def normal(x, loc=0., sca=1., sig=None):
    """
    return normal(*args, **kwargs)


# --------------------------------------------------------------------


def laplace(x, loc=0., sca=1., sig=None):
    """
        Laplace probability density function (pdf).


        Definition
        ----------
        def laplace(x, loc=0., sca=1., sig=None):


        Input
        -----
        x          array_like quantiles


        Optional Input
        --------------
        loc        location
        sca        scale
        sig        standard deviation, overwrites scale


        Output
        ------
        Laplace pdf at x


        Examples
        --------
        >>> print(str(laplace(0.)))
        0.5

        >>> print(str(laplace(1.) - 0.5/np.e))
        0.0

        >>> print(str(laplace(0., 0., 2.)))
        0.25

        >>> print(str(laplace(0., 2., 2.) - 0.25/np.e))
        0.0

        >>> print(np.allclose(laplace(1., 2., 2.), laplace((1.-2.)/2.)/2.))
        True


        History
        -------
        Written,  MC, May 2016
    """

    if sig is not None: sca = sig/np.sqrt(2.)
    return laplace01((x-loc)/sca)/sca


def laplace01(x):
    """
        Laplace probability density function (pdf) with at location zero and unit scale.


        Definition
        ----------
        def laplace01(x):


        Input
        -----
        x          array_like quantiles


        Optional Input
        --------------
        None


        Output
        ------
        Laplace pdf with loc=0 and sca=1 at x


        Examples
        --------
        >>> print(str(laplace01(0.)))
        0.5

        >>> print(str(laplace01(1.) - 0.5/np.e))
        0.0


        History
        -------
        Written,  MC, May 2016
    """

    return 0.5 * np.exp(-np.abs(x))


# --------------------------------------------------------------------


def multigauss(*args, **kwargs):
    """
        Wrapper for multinormal

        def multinormal(x, loc=0., sca=1., sig=None):
    """
    return multinormal(*args, **kwargs)


# --------------------------------------------------------------------


def multinorm(*args, **kwargs):
    """
        Wrapper for multinormal

        def multinormal(x, loc=0., sca=1., sig=None):
    """
    return multinormal(*args, **kwargs)


# --------------------------------------------------------------------


def multinormal(x, loc=0., sca=1., cov=None):
    """
        Multivariate normal (Gauss) probability density function (pdf).


        Definition
        ----------
        def multinormal(x, loc=0., sca=1., sig=None):


        Input
        -----
        x          array_like quantiles


        Optional Input
        --------------
        loc        array_like location
        sca        array_like scale
                   Scalar: single variance for all dimensions
                   1D: variances for all dimensions
                   2D: covariance matrix
        cov        array_like standard deviation, overwrites scale


        Output
        ------
        Multivariate normal (Gauss) pdf at x


        Examples
        --------
        >>> print(np.allclose(multinormal(0.), 1./np.sqrt(2.*np.pi)))
        True

        >>> print(np.allclose(multinormal(1.), 1./np.sqrt(2.*np.pi*np.e)))
        True

        >>> print(np.allclose(multinormal(0., 0., 2.), 1./np.sqrt(2.*np.pi)/np.sqrt(2.)))
        True

        >>> print(np.allclose(multinormal(0., np.sqrt(2.), 1.)*np.sqrt(2.*np.pi), 1./np.e))
        True

        >>> print(np.allclose(multinormal(1., 2., 3.), multinormal((1.-2.)/np.sqrt(3.))/np.sqrt(3.)))
        True


        History
        -------
        Written,  MC, Dec 2017
    """

    if cov is not None: sca = cov
    k = np.size(x)
    if k==1:
        ix = np.array([x])
    else:
        ix = np.array(x)
    l = np.size(loc)
    if l==1:
        iloc = np.array([loc])
    else:
        iloc = np.array(loc)
    if np.ndim(sca) == 0:
        icov = np.diag(np.full(k, sca)) # np.full((k,k), sca)
    elif np.ndim(sca) == 1:
        icov = np.diag(sca)
    elif np.ndim(sca) == 2:
        icov = sca
    detcov = np.linalg.det(icov)
    assert detcov != 0., 'Degenerate case: determinant of covariance matrix == 0.'
    part1 = np.sqrt((2.*np.pi)**k * detcov)
    # transpose different to matrix formula because of broadcasting rules in Python
    # x (ndim,) or (nsample,ndim); loc scalar or (ndim,); cov (ndim,ndim)
    part2 = -0.5 * np.dot(np.dot((ix-iloc).T, np.linalg.inv(icov)), (ix-iloc))
    return np.exp(part2)/part1


# --------------------------------------------------------------------


def norm(*args, **kwargs):
    """
        Wrapper for normal

        def normal(x, loc=0., sca=1., sig=None):
    """
    return normal(*args, **kwargs)


# --------------------------------------------------------------------


def normal(x, loc=0., sca=1., sig=None):
    """
        Normal (Gauss) probability density function (pdf).


        Definition
        ----------
        def normal(x, loc=0., sca=1., sig=None):


        Input
        -----
        x          array_like quantiles


        Optional Input
        --------------
        loc        location
        sca        scale
        sig        standard deviation, overwrites scale


        Output
        ------
        Normal (Gauss) pdf at x


        Examples
        --------
        >>> print(np.allclose(normal(0.), 1./np.sqrt(2.*np.pi)))
        True

        >>> print(np.allclose(normal(1.), 1./np.sqrt(2.*np.pi*np.e)))
        True

        >>> print(np.allclose(normal(0., 0., 2.), 0.5/np.sqrt(2.*np.pi)))
        True

        >>> print(np.allclose(normal(0., np.sqrt(2.), 1.)*np.sqrt(2.*np.pi), 1./np.e))
        True

        >>> print(np.allclose(normal(1., 2., 3.), normal((1.-2.)/3.)/3.))
        True


        History
        -------
        Written,  MC, May 2016
    """

    if sig is not None: sca = sig
    return normal01((x-loc)/sca)/sca


def normal01(x):
    """
        Normal (Gauss) probability density function (pdf) at location zero and unit scale.


        Definition
        ----------
        def normal01(x):


        Input
        -----
        x          array_like quantiles


        Optional Input
        --------------
        None


        Output
        ------
        Normal pdf with loc=0 and sca=1 at x


        Examples
        --------
        >>> print(np.allclose(normal01(0.), 1./np.sqrt(2.*np.pi)))
        True

        >>> print(np.allclose(normal01(1.), 1./np.sqrt(2.*np.pi*np.e)))
        True


        History
        -------
        Written,  MC, May 2016
    """

    return 1./np.sqrt(2.*np.pi) * np.exp(-0.5*x**2)


# --------------------------------------------------------------------


def sep(x, loc=0., sca=1., xi=1., beta=0., sig=None):
    """
        The skew exponential power distribution with given location, scale, skewness, and kurtosis.


        Definition
        ----------
        def sep(x, loc=0., sca=1., xi=1., beta=0., sig=None):


        Input
        -----
        x          array_like quantiles


        Optional Input
        --------------
        loc        location
        sca        scale
        xi         skewness parameter
        beta       kurtosis parameter
        sig        standard deviation, overwrites scale


        Output
        ------
        Skew exponential power pdf at x


        Literature
        ----------
        Fernandez C & Steel M (1998) On Bayesian modeling of fat tails and skewness.
            Journal of the American Statistical Association 93, 359-371.
        Schoups G & Vrugt JA (2010) A formal likelihood function for parameter and predictive
            inference of hydrologic models with correlated, heteroscedastic, and non-Gaussian errors.
            Water Resources Research 46, W10531.


        Examples
        --------
        >>> print(np.allclose(sep(1., 2., 2., 0.5, 2.), sep((1.-2.)/2., xi=0.5, beta=2.)/2.))
        True

        >>> print(np.allclose(sep(1.3, 0., 1., 1., 0.), normal(1.3, 0., 1.)))
        True

        >>> print(np.allclose(sep(1.3, 0., 1., 1., 1.), laplace(1.3, 0., 1./np.sqrt(2.))))
        True

        >>> print(np.allclose(sep(1.3, 0., np.sqrt(2.), 1., 1.), laplace(1.3, 0., 1.)))
        True


        History
        -------
        Written,  MC, May 2016
    """

    if sig is not None: sca = sig
    return sep01((x-loc)/sca, xi, beta)/sca


def sep01(x, xi=1., beta=0.):
    """
        The skew exponential power distribution with given skewness and kurtosis parameters,
        location zero and unit scale.


        Definition
        ----------
        def sep01(x, xi=1., beta=0.):


        Input
        -----
        x          array_like quantiles


        Optional Input
        --------------
        xi         skewness parameter
        beta       kurtosis parameter


        Output
        ------
        Skew exponential power pdf with loc=0, sca=1 at x


        Literature
        ----------
        Fernandez C & Steel M (1998) On Bayesian modeling of fat tails and skewness.
            Journal of the American Statistical Association 93, 359-371.
        Schoups G & Vrugt JA (2010) A formal likelihood function for parameter and predictive
            inference of hydrologic models with correlated, heteroscedastic, and non-Gaussian errors.
            Water Resources Research 46, W10531.


        History
        -------
        Written,  MC, May 2016
    """

    mu  = sep01_fs_mean(xi, beta)
    sig = sep01_fs_std(xi, beta)
    z   = mu + sig*x

    return sig * sep01_fs(z, xi, beta)


# --------------------------------------------------------------------


def sep_fs(x, loc=0., sca=1., xi=1., beta=0., sig=None):
    """
        The skew exponential power distribution Fernandez C & Steel M (1998)
        with given location, scale, skewness, and kurtosis.


        Definition
        ----------
        def sep_fs(x, loc=0., sca=1., xi=1., beta=0., sig=None):


        Input
        -----
        x          array_like quantiles


        Optional Input
        --------------
        loc        location
        sca        scale
        xi         skewness parameter
        beta       kurtosis parameter
        sig        standard deviation, overwrites scale


        Output
        ------
        Skew exponential power pdf at x after Fernandez and Steel (1998)


        Examples
        --------
        >>> print(np.allclose(sep(1., 2., 2., 0.5, 2.), sep((1.-2.)/2., xi=0.5, beta=2.)/2.))
        True

        >>> print(np.allclose(sep(1.3, 0., 1., 1., 0.), normal(1.3, 0., 1.)))
        True

        >>> print(np.allclose(sep(1.3, 0., 1., 1., 1.), laplace(1.3, 0., 1./np.sqrt(2.))))
        True

        >>> print(np.allclose(sep(1.3, 0., np.sqrt(2.), 1., 1.), laplace(1.3, 0., 1.)))
        True


        History
        -------
        Written,  MC, May 2016
    """

    if sig is not None: sca = sig
    return sep01_fs((x-loc)/sca, xi, beta)/sca


def sep01_fs(x, xi=1., beta=0.):
    """
        The skew exponential power distribution after Fernandez and Steel (1998)
        with location zero and unit scale.

        Definition
        ----------
        def sep01_fs(x, xi=1., beta=0.):


        Input
        -----
        x          array_like quantiles


        Optional Input
        --------------
        xi         skewness parameter
        beta       kurtosis parameter


        Output
        ------
        Skew exponential power pdf with loc=0, sca=1 at x after Fernandez and Steel (1998)


        Literature
        ----------
        Fernandez C & Steel M (1998) On Bayesian modeling of fat tails and skewness.
            Journal of the American Statistical Association 93, 359-371.


        History
        -------
        Written,  MC, May 2016
    """

    alpha = np.where(x<0.0, xi, 1./xi)
    if not np.iterable(x): alpha = float(alpha)

    return 2.0/(xi+1./xi) * ep01(alpha*x, beta)


# --------------------------------------------------------------------


def sep_fs_mean(loc=0., sca=1., xi=1., beta=0., sig=None):
    """
        Mean of skew exponential power distribution with given skewness and kurtosis parameters,
        location and scale.


        Definition
        ----------
        def sep_fs_mean(loc=0., sca=1., xi=1., beta=0., sig=None):


        Optional Input
        --------------
        loc        location
        sca        scale
        xi         skewness parameter
        beta       kurtosis parameter
        sig        standard deviation, overwrites scale


        Output
        ------
        Mean of skew exponential power pdf at x


        History
        -------
        Written,  MC, May 2016
    """

    if sig is not None: sca = sig
    return sep01_fs_mean(xi, beta) * sca + loc


def sep01_fs_mean(xi=1., beta=0.):
    """
        Mean of skew exponential power distribution with given skewness and kurtosis parameters,
        location zero and unit scale.


        Definition
        ----------
        def sep01_fs_mean(xi=1., beta=0.):


        Optional Input
        --------------
        xi         skewness parameter
        beta       kurtosis parameter


        Output
        ------
        Mean of skew exponential power pdf with loc=0, sca=1 at x


        Literature
        ----------
        Fernandez C & Steel M (1998) On Bayesian modeling of fat tails and skewness.
            Journal of the American Statistical Association 93, 359-371.
        Schoups G & Vrugt JA (2010) A formal likelihood function for parameter and predictive
            inference of hydrologic models with correlated, heteroscedastic, and non-Gaussian errors.
            Water Resources Research 46, W10531.

        History
        -------
        Written,  MC, May 2016
    """
    import scipy.special as ss

    if beta != -1.0:
        b1 = 0.5*(1.0 + beta)
        b2 =      1.0 + beta
        b3 = 1.5*(1.0 + beta)
        g1 = ss.gamma(b1)
        g2 = ss.gamma(b2)
        g3 = ss.gamma(b3)
        # -> sqrt(3/4)
        M1 = g2 / np.sqrt(g3*g1)
    else:
        M1 = np.sqrt(0.75)

    if xi != 1:
        return M1*(xi-1./xi)
    else:
        return 0.0


# --------------------------------------------------------------------


def sep_fs_std(sca=1., xi=1., beta=0., sig=None):
    """
        Standard deviation of skew exponential power distribution with given skewness and kurtosis parameters,
        location and scale.


        Definition
        ----------
        def sep_fs_std(sca=1., xi=1., beta=0., sig=None):


        Optional Input
        --------------
        sca        scale
        xi         skewness parameter
        beta       kurtosis parameter
        sig        standard deviation, overwrites scale


        Output
        ------
        Standard deviation of skew exponential power pdf at x


        Literature
        ----------
        Fernandez C & Steel M (1998) On Bayesian modeling of fat tails and skewness.
            Journal of the American Statistical Association 93, 359-371.
        Schoups G & Vrugt JA (2010) A formal likelihood function for parameter and predictive
            inference of hydrologic models with correlated, heteroscedastic, and non-Gaussian errors.
            Water Resources Research 46, W10531.


        History
        -------
        Written,  MC, May 2016
    """

    if sig is not None: sca = sig
    return sep01_fs_std(xi, beta) * sca


def sep01_fs_std(xi=1., beta=0.):
    """
        Standard deviation of skew exponential power distribution with given skewness and kurtosis parameters,
        location zero and unit scale.


        Definition
        ----------
        def sep01_fs_std(xi=1., beta=0.):


        Optional Input
        --------------
        xi         skewness parameter
        beta       kurtosis parameter


        Output
        ------
        Standard deviation of skew exponential power pdf with loc=0, sca=1 at x


        Literature
        ----------
        Fernandez C & Steel M (1998) On Bayesian modeling of fat tails and skewness.
            Journal of the American Statistical Association 93, 359-371.
        Schoups G & Vrugt JA (2010) A formal likelihood function for parameter and predictive
            inference of hydrologic models with correlated, heteroscedastic, and non-Gaussian errors.
            Water Resources Research 46, W10531.


        History
        -------
        Written,  MC, May 2016
    """

    mu  = sep01_fs_mean(xi, beta)
    if xi != 1.:
        M1 = mu / (xi-1./xi)
    else:
        M1 = 0.
    M2  = 1.0
    var = (M2-M1**2)*(xi**2+1./xi**2) + 2.0*M1**2 - M2
    if var > 0.0:
        sig  = np.sqrt(var)
    else:
        sig  = 0.0

    return sig


# --------------------------------------------------------------------


def st(x, nu, loc=0., sca=1., xi=1., sig=None):
    """
        The standardised skewed Student t distribution with given degrees of freedom, location, scale, and skewness.


        Definition
        ----------
        def st(x, nu, loc=0., sca=1., xi=1., sig=None):


        Input
        -----
        x          array_like quantiles
        nu         degrees of freedom


        Optional Input
        --------------
        loc        location
        sca        scale
        xi         skewness parameter
        sig        standard deviation, overwrites scale


        Output
        ------
        Skewed Student t pdf at x


        Examples
        --------
        >>> print(np.allclose(st(1., 3., 2., 0.5, 2.), st((1.-2.)/0.5, 3., xi=2.)/0.5))
        True


        History
        -------
        Written,  MC, May 2016
    """

    if sig is not None: sca = sig * np.sqrt((nu-2.)/nu)
    return st01((x-loc)/sca, nu, xi)/sca


def st01(x, nu, xi=1.):
    """
        The standardised skewed Student t distribution with given degrees of freedom and skewness,
        location zero and unit scale.


        Definition
        ----------
        def st01(x, nu, xi=1.):


        Input
        -----
        x          array_like quantiles
        nu         degrees of freedom


        Optional Input
        --------------
        xi       skewness parameter


        Output
        ------
        Skewed Student t pdf with loc=0 and sca=1 at x


        Literature
        ----------
        Fernandez C & Steel M (1998) On Bayesian modeling of fat tails and skewness.
            Journal of the American Statistical Association 93, 359-371.
        Schoups G & Vrugt JA (2010) A formal likelihood function for parameter and predictive
            inference of hydrologic models with correlated, heteroscedastic, and non-Gaussian errors.
            Water Resources Research 46, W10531.


        History
        -------
        Written,  MC, May 2016
    """

    mu  = st01_fs_mean(nu, xi)
    sca = st01_fs_std(nu, xi) * np.sqrt((nu-2.)/nu)
    z   = mu + sca*x

    return sca * st01_fs(z, nu, xi)


# --------------------------------------------------------------------


def st_fs(x, nu, loc=0., sca=1., xi=1., sig=None):
    """
        The skewed Student t distribution with given degrees of freedom, location, scale, and skewness.


        Definition
        ----------
        def st(x, nu, loc=0., sca=1., xi=1., sig=None):


        Input
        -----
        x          array_like quantiles
        nu         degrees of freedom


        Optional Input
        --------------
        loc        location
        sca        scale
        xi         skewness parameter
        sig        standard deviation, overwrites scale


        Output
        ------
        Skewed Student t pdf at x


        Literature
        ----------
        Fernandez C & Steel M (1998) On Bayesian modeling of fat tails and skewness.
            Journal of the American Statistical Association 93, 359-371.


        History
        -------
        Written,  MC, May 2016
    """

    if sig is not None: sca = sig * np.sqrt((nu-2.)/nu)
    return st01_fs((x-loc)/sca, nu, xi)/sca


def st01_fs(x, nu, xi=1.):
    """
        The skewed Student t distribution with given degrees of freedom and skewness,
        location zero and unit scale.

        If xi is not 1 then mean is not zero and standard deviation is not 1.


        Definition
        ----------
        def st01_fs(x, nu, xi=1.):


        Input
        -----
        x          array_like quantiles
        nu         degrees of freedom


        Optional Input
        --------------
        xi         skewness parameter


        Output
        ------
        Skewed Student t pdf with loc=0 and sca=1 at x


        Literature
        ----------
        Fernandez C & Steel M (1998) On Bayesian modeling of fat tails and skewness.
            Journal of the American Statistical Association 93, 359-371.


        History
        -------
        Written,  MC, May 2016
    """

    alpha = np.where(x<0.0, xi, 1./xi)
    if not np.iterable(x): alpha = float(alpha)

    return 2.0/(xi+1./xi) * t01(alpha*x, nu)


def st_fs_mean(nu, loc=0., sca=1., xi=1., sig=None):
    """
        The mean of the skewed Student t distribution with given degrees of freedom,
        location, scale, and skewness.


        Definition
        ----------
        def st_fs_mean(nu, loc=0., sca=1., xi=1., sig=None):


        Input
        -----
        nu         degrees of freedom


        Optional Input
        --------------
        loc        location
        sca        scale
        xi         skewness parameter
        sig        standard deviation, overwrites scale


        Output
        ------
        Mean of skewed Student t pdf


        Literature
        ----------
        Fernandez C & Steel M (1998) On Bayesian modeling of fat tails and skewness.
            Journal of the American Statistical Association 93, 359-371.


        History
        -------
        Written,  MC, May 2016
    """

    if sig is not None: sca = sig * np.sqrt((nu-2.)/nu)
    return st01_fs_mean(nu, xi) * sca + loc


def st01_fs_mean(nu, xi=1.):
    """
        The mean of the skewed Student t distribution with given degrees of freedom and skewness,
        location zero and unit scale.


        Definition
        ----------
        def st01_fs_mean(nu, xi=1.):


        Input
        -----
        nu         degrees of freedom


        Optional Input
        --------------
        xi         skewness parameter


        Output
        ------
        Mean of skewed Student t pdf with loc=0 and sca=1 at x


        Literature
        ----------
        Scharnagl B et al. (2015) Inverse modelling of in situ soil water dynamics:
            accounting for heteroscedastic, autocorrelated, and non-Gaussian distributed residuals.
            Hydrology And Earth System Sciences Discussions 12, 2155-2199.


        History
        -------
        Written,  MC, May 2016
    """
    import scipy.special as ss

    if nu <= 1.:
        return np.inf
    else:
        mu  = 2.0 * (xi-1./xi) * ss.gamma(0.5*(nu+1.0)) / ss.gamma(0.5*nu)
        mu *= (nu-2.0)/(nu-1.0) * nu/(nu-2.) / np.sqrt(np.pi*nu)

    return mu


def st_fs_std(nu, sca=1., xi=1., sig=None):
    """
        The standard deviation of the skewed Student t distribution with given degrees of freedom,
        scale, and skewness.


        Definition
        ----------
        def st_fs_std(nu, sca=1., xi=1., sig=None):


        Input
        -----
        nu         degrees of freedom


        Optional Input
        --------------
        sca        scale
        xi         skewness parameter
        sig        standard deviation, overwrites scale


        Output
        ------
        Standard deviation of skewed Student t pdf


        Literature
        ----------
        Fernandez C & Steel M (1998) On Bayesian modeling of fat tails and skewness.
            Journal of the American Statistical Association 93, 359-371.


        History
        -------
        Written,  MC, May 2016
    """

    if sig is not None: sca = sig * np.sqrt((nu-2.)/nu)
    return st01_fs_std(nu, xi) * sca


def st01_fs_std(nu, xi=1.):
    """
        The standard deviation of the skewed Student t distribution with given degrees of freedom and skewness,
        and unit scale.


        Definition
        ----------
        def st01_fs_std(nu, xi=1.):


        Input
        -----
        nu         degrees of freedom


        Optional Input
        --------------
        xi       skewness parameter


        Output
        ------
        Standard deviation of skewed Student t pdf with sca=1 at x


        Literature
        ----------
        Fernandez C & Steel M (1998) On Bayesian modeling of fat tails and skewness.
            Journal of the American Statistical Association 93, 359-371.


        History
        -------
        Written,  MC, May 2016
    """

    if nu <= 2.:
        return np.inf
    else:
        mu = st01_fs_mean(nu, xi)
        if xi != 1.:
            M1 = mu / (xi-1./xi)
        else:
            M1 = 0.
        M2  = nu/(nu-2.)
        var = (M2-M1**2)*(xi**2+1./xi**2)+2.*M1**2-M2
        if var > 0.0:
            return np.sqrt(var)
        else:
            return 0.0


# --------------------------------------------------------------------


def t(x, nu, loc=0., sca=1., sig=None):
    """
        The Student t distribution with given degrees of freedom, location, scale, and skewness.


        Definition
        ----------
        def t(x, nu, loc=0., sca=1., sig=None):


        Input
        -----
        x          array_like quantiles
        nu         degrees of freedom


        Optional Input
        --------------
        loc        location
        sca        scale
        sig        standard deviation, overwrites scale


        Output
        ------
        Student t pdf at x


        Examples
        --------
        >>> print(np.allclose(st(1., 3., 2., 0.5, 2.), st((1.-2.)/0.5, 3., xi=2.)/0.5))
        True


        History
        -------
        Written,  MC, May 2016
    """

    if sig is not None: sca = sig * np.sqrt((nu-2.)/nu)
    return t01((x-loc)/sca, nu)/sca


def t01(x, nu):
    """
        The Student t distribution with given degrees of freedom and skewness,
        location zero and unit scale.


        Definition
        ----------
        def t01(x, nu):


        Input
        -----
        x          array_like quantiles
        nu         degrees of freedom


        Output
        ------
        Student t pdf with loc=0 and sca=1 at x


        History
        -------
        Written,  MC, May 2016
    """
    import scipy.special as ss

    c = ss.gamma(0.5*(nu+1.0)) / (ss.gamma(0.5*nu) * np.sqrt(np.pi*nu))
    return c * (1.0 + x**2/nu)**(-0.5*(nu+1.0))


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # import numpy as np
    # import scipy.special as spec
    # import scipy.stats as sp
    # import jams

    # nn = 10000000
    # xx = np.arange(nn)/float(nn) * 30.*100. - 15.*100.

    # loc=1.1
    # sca=2.2
    # sig=sca
    # xi=3.3
    # beta=0.5
    # nu=4.4

    # dx = xx[2]-xx[1]

    # __all__ = ['ep', 'exponential', 'laplace', 'normal', 'sep', 'st', 't']

    # print('loc/sca/sig/sig**2/xi/beta/nu:', loc, sca, sig, sig**2, xi, beta, nu)
    # for dd in __all__:
    #     # use sca
    #     if dd == 'ep':
    #         pdf = jams.distributions.ep(xx, loc, sca, beta)
    #     elif dd == 'exponential':
    #         pdf = jams.distributions.exponential(xx, loc, sca)
    #     elif dd == 'laplace':
    #         pdf = jams.distributions.laplace(xx, loc, sca)
    #     elif dd == 'normal':
    #         pdf = jams.distributions.normal(xx, loc, sca)
    #     elif dd == 'sep':
    #         pdf = jams.distributions.sep(xx, loc, sca, xi, beta)
    #     elif dd == 'st':
    #         pdf = jams.distributions.st(xx, nu, loc, sca, xi)
    #     elif dd == 't':
    #         pdf = jams.distributions.t(xx, nu, loc, sca)

    #     dmean = np.sum(xx * pdf * dx)
    #     dvar  = np.sum((xx-dmean)**2 * pdf * dx)
    #     dstd  = np.sqrt(dvar)
    #     dskew = np.sum(((xx-dmean)/dstd)**3 * pdf * dx)
    #     dkurt = np.sum(((xx-dmean)/dstd)**4 * pdf * dx)
    #     print(dd, ':', dmean, dstd, dvar, dskew, dkurt)

    #     # use sig
    #     if dd == 'ep':
    #         pdf = jams.distributions.ep(xx, loc, beta=beta, sig=sig)
    #     elif dd == 'exponential':
    #         pdf = jams.distributions.exponential(xx, loc, sig=sig)
    #     elif dd == 'laplace':
    #         pdf = jams.distributions.laplace(xx, loc, sig=sig)
    #     elif dd == 'normal':
    #         pdf = jams.distributions.normal(xx, loc, sig=sig)
    #     elif dd == 'sep':
    #         pdf = jams.distributions.sep(xx, loc, xi=xi, beta=beta, sig=sig)
    #     elif dd == 'st':
    #         pdf = jams.distributions.st(xx, nu, loc, xi=xi, sig=sig)
    #     elif dd == 't':
    #         pdf = jams.distributions.t(xx, nu, loc, sig=sig)

    #     dmean = np.sum(xx * pdf * dx)
    #     dvar  = np.sum((xx-dmean)**2 * pdf * dx)
    #     dstd  = np.sqrt(dvar)
    #     dskew = np.sum(((xx-dmean)/dstd)**3 * pdf * dx)
    #     dkurt = np.sum(((xx-dmean)/dstd)**4 * pdf * dx)
    #     print(dd, ':', dmean, dstd, dvar, dskew, dkurt)

    #     # # xx   = xx[xx>0.]
    #     # # pdf  = jams.distributions.t(xx, nu, loc, sca)
    #     # # M1   = 2.* np.sum(xx * pdf * dx)
    #     # # M2   = nu/(nu-2.) * sca**2
    #     # # dvarfs = (M2-M1**2)*(xi**2+1./xi**2)+2.*M1**2-M2
    #     # # # print('Fernandez var:', dvarfs)

    #     # mu_shmc  = 2.0 * (xi-1./xi) * spec.gamma(0.5*(nu+1.0)) / spec.gamma(0.5*nu)
    #     # mu_shmc *= (nu-2.0)/(nu-1.0)
    #     # mu_shmc *= nu/(nu-2.) / np.sqrt(np.pi*nu)
    #     # mu_shmc *= sca

    #     # # var_shmc = -mu_shmc**2 + (xi**3+1./xi**3)/(xi+1./xi)
    #     # # # print('mu_shmc:', mu_shmc, var_shmc)

    #     # if xi != 1.:
    #     #     M1 = mu_shmc / (xi-1./xi)
    #     # else:
    #     #     M1 = 0.
    #     # M2   = nu/(nu-2.) * sca**2
    #     # dvar_mc = (M2-M1**2)*(xi**2+1./xi**2)+2.*M1**2-M2
    #     # print('dvar_mc:', mu_shmc, dvar_mc)



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
"""
    Defines common error measures and discharge signatures


    Error measures
    --------------
    def bias(obs,mod):      bias
                            mean(obs) - mean(mod)
    def mae(obs,mod):       mean absolute error
                            mean(abs(obs - mod))
    def mse(obs,mod):       mean square error
                            mean((obs - mod)**2)
    def rmse(obs,mod):      root mean square error
                            sqrt(mean((obs - mod)**2))
    def nse(obs,mod):       Nash-Sutcliffe efficiency
                            1 - sum((obs - mod)**2) / sum((obs - mean(obs))**2)
    def kge(obs,mod):       Kling-Gupta efficiency
                            1 - sqrt((1-r)**2 + (1-a)**2 + (1-b)**2),
                            where r is the Pearson correlation of obs and mod,
                                  a is mean(mod) / mean(obs), and
                                  b is std(mod) / std(obs)
    def pearson(obs,mod):   Pearson's correlation coefficient
                            mean((obs-mean(obs))/stddev(obs) * (mod-mean(mod))/stddev(mod))

    Input
    -----
    obs         ND-array
    mod         ND-array
    quantiles   Scalar or 1D array_like percentages of exceedance


    Output
    ------
    Measure calculated along the first axis.



    Discharge signatures
    --------------------
    def autocorrelation(dat, lags):
                            Autocorrelation of a data series at given lags.
    def flowdurationcurve(dat, quantiles=None, concavity_index=False,
                          mid_segment_slope=False, mhigh_segment_volume=False,
                          high_segment_volume=False, low_segment_volume=False):
                            Flow duration curves for a given data vector. The Flow duration curve at a
                            certain quantile x is the data point p where x% of the data points are above the value p.
                            Optionally, can be calculated:
                                the concavity index CI can be calculated [Zhang2014].
                                the FDC mid-segment slope [Shafii et. al 2014]
                                the FDC medium high-segment volume [Shafii et. al 2014]
                                the FDC high-segment volume [Shafii et. al 2014]
                                the FDC low-segment volume [Shafii et. al 2014]
    def limbdensities(dat):
                            Rising and declinging limb densities,
                            which are the duration of the data increase (decrease) divided by the number of peaks.
    def maximummonthlyflow(date, dat):
                            Maximum of average flows per month
    def moments(dat, mean_data=False, stddev_data=False, median_data=False,
                            Moments of data and log-transformed data
                            Returns several moments of data series given, i.e.
                                * mean               of data
                                * standard deviation of data
                                * median             of data
                                * maximum/ peak      of data
                                * mean               of log-transformed data
                                * standard deviation of log-transformed data
                                * median             of log-transformed data
                                * maximum/ peak      of log-transformed data
     def peakdistribution(dat, quantiles=None, slope_peak_distribution=False):
                            Calculates the peak distribution.
                            Optionally, the slope of the peak distribution can be returned.


    Input
    -----
    dat         1D array_like
    quantiles   Scalar or 1D array_like percentages of exceedance


    Output
    ------
    Discharge signature.


    Examples
    --------
    >>> # Create some data
    >>> obs = np.array([12.7867, 13.465, 14.1433, 15.3733, 16.6033])
    >>> mod = np.array([12.8087, 13.151, 14.3741, 16.2302, 17.9433])

    >>> # bias
    >>> print(np.round(bias(obs, mod),2))
    -0.43

    >>> print(np.round(mae(obs, mod),2))
    0.55

    >>> print(np.round(mse(obs, mod),2))
    0.54

    >>> print(np.round(rmse(obs, mod),2))
    0.73

    >>> print(np.round(nse(obs, mod),2))
    0.71

    >>> print(np.round(kge(obs, mod),2))
    0.58

    >>> print(np.round(pear2(obs, mod),2))
    0.99

    >>> print(np.round(confint(obs, p=0.95)))
    [ 13.  16.]


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2016 Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written, MC, May 2016
"""

from .quality_assess import bias, mae, mse, rmse, nse, kge, pearson
from .signatures     import autocorrelation, flowdurationcurve, limbdensities
from .signatures     import maximummonthlyflow, moments, peakdistribution

# Information
__author__   = "Matthias Cuntz"
__version__  = '1.0'
__revision__ = "Revision: 2419"
__date__     = 'Date: 16.05.2016'



#!/usr/bin/env python
"""
This is the unittest for closest module.

python -m unittest -v tests/test_closest.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_closest.py

"""
import unittest


class TestClosest(unittest.TestCase):
    """
    Tests for closest.py
    """

    def test_closest(self):
        import numpy as np
        import pandas as pd
        from pyjams import closest

        nn = 51

        # 1D
        arr = np.arange(nn) / float(nn - 1) * 5.
        assert closest(arr, 3.125) == 31
        assert closest(arr, 3.125, value=True) == 3.1

        df = pd.Series(arr)
        df.index = pd.date_range('2022-01-01', periods=nn)
        assert closest(df, 3.125) == 31
        assert closest(df, 3.125, value=True) == 3.1

        # 2D
        arr2 = np.arange(nn).reshape((17, 3)) / float(nn - 1) * 5.
        assert closest(arr2, 3.125, value=True) == 3.1
        ii = np.unravel_index(closest(arr2, 3.125), arr2.shape)
        self.assertEqual(ii, (10, 1))
        assert arr2[ii] == 3.1

        df2 = pd.DataFrame(arr2)
        df2.index = pd.date_range('2022-01-01', periods=17)
        assert closest(df2, 3.125, value=True) == 3.1
        ii = np.unravel_index(closest(df2, 3.125), df2.shape)
        self.assertEqual(ii, (10, 1))
        assert df2.iloc[ii] == 3.1


if __name__ == "__main__":
    unittest.main()
